### ms-scmr.idl
"""
RPC definitions for the following interfaces:
- svcctl (v2.0): 367ABB81-9844-35F1-AD32-98F038001003
This file is auto-generated by midl-to-scapy, do not modify.
"""

from enum import IntEnum
import uuid

from scapy.fields import StrFixedLenField
from scapy.layers.dcerpc import (
    NDRPacket,
    DceRpcOp,
    NDRByteField,
    NDRConfPacketListField,
    NDRConfStrLenField,
    NDRConfVarStrLenField,
    NDRConfVarStrLenFieldUtf16,
    NDRConfVarStrNullField,
    NDRConfVarStrNullFieldUtf16,
    NDRContextHandle,
    NDRFullPointerField,
    NDRIntEnumField,
    NDRIntField,
    NDRLongField,
    NDRPacketField,
    NDRShortField,
    NDRSignedIntField,
    NDRUnionField,
    register_dcerpc_interface,
)


class RCloseServiceHandle_Request(NDRPacket):
    fields_desc = [NDRPacketField("hSCObject", NDRContextHandle(), NDRContextHandle)]


class RCloseServiceHandle_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCObject", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class LPSERVICE_STATUS(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwCurrentState", 0),
        NDRIntField("dwControlsAccepted", 0),
        NDRIntField("dwWin32ExitCode", 0),
        NDRIntField("dwServiceSpecificExitCode", 0),
        NDRIntField("dwCheckPoint", 0),
        NDRIntField("dwWaitHint", 0),
    ]


class RControlService_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwControl", 0),
    ]


class RControlService_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("lpServiceStatus", LPSERVICE_STATUS(), LPSERVICE_STATUS),
        NDRIntField("status", 0),
    ]


class RDeleteService_Request(NDRPacket):
    fields_desc = [NDRPacketField("hService", NDRContextHandle(), NDRContextHandle)]


class RDeleteService_Response(NDRPacket):
    fields_desc = [NDRIntField("status", 0)]


class RLockServiceDatabase_Request(NDRPacket):
    fields_desc = [NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle)]


class RLockServiceDatabase_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("lpLock", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class RQueryServiceObjectSecurity_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwSecurityInformation", 0),
        NDRIntField("cbBufSize", 0),
    ]


class RQueryServiceObjectSecurity_Response(NDRPacket):
    fields_desc = [
        NDRConfStrLenField(
            "lpSecurityDescriptor", "", size_is=lambda pkt: pkt.cbBufSize
        ),
        NDRIntField("pcbBytesNeeded", 0),
        NDRIntField("status", 0),
    ]


class RSetServiceObjectSecurity_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwSecurityInformation", 0),
        NDRConfStrLenField(
            "lpSecurityDescriptor", "", size_is=lambda pkt: pkt.cbBufSize
        ),
        NDRIntField("cbBufSize", None, size_of="lpSecurityDescriptor"),
    ]


class RSetServiceObjectSecurity_Response(NDRPacket):
    fields_desc = [NDRIntField("status", 0)]


class RQueryServiceStatus_Request(NDRPacket):
    fields_desc = [NDRPacketField("hService", NDRContextHandle(), NDRContextHandle)]


class RQueryServiceStatus_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("lpServiceStatus", LPSERVICE_STATUS(), LPSERVICE_STATUS),
        NDRIntField("status", 0),
    ]


class RSetServiceStatus_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hServiceStatus", NDRContextHandle(), NDRContextHandle),
        NDRPacketField("lpServiceStatus", LPSERVICE_STATUS(), LPSERVICE_STATUS),
    ]


class RSetServiceStatus_Response(NDRPacket):
    fields_desc = [NDRIntField("status", 0)]


class RUnlockServiceDatabase_Request(NDRPacket):
    fields_desc = [NDRPacketField("Lock", NDRContextHandle(), NDRContextHandle)]


class RUnlockServiceDatabase_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("Lock", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class RNotifyBootConfigStatus_Request(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpMachineName", "")),
        NDRIntField("BootAcceptable", 0),
    ]


class RNotifyBootConfigStatus_Response(NDRPacket):
    fields_desc = [NDRIntField("status", 0)]


class RChangeServiceConfigW_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwStartType", 0),
        NDRIntField("dwErrorControl", 0),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpBinaryPathName", "")),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpLoadOrderGroup", "")),
        NDRFullPointerField(NDRIntField("lpdwTagId", 0)),
        NDRFullPointerField(
            NDRConfStrLenField(
                "lpDependencies", "", size_is=lambda pkt: pkt.dwDependSize
            )
        ),
        NDRIntField("dwDependSize", None, size_of="lpDependencies"),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpServiceStartName", "")),
        NDRFullPointerField(
            NDRConfStrLenField("lpPassword", "", size_is=lambda pkt: pkt.dwPwSize)
        ),
        NDRIntField("dwPwSize", None, size_of="lpPassword"),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpDisplayName", "")),
    ]


class RChangeServiceConfigW_Response(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRIntField("lpdwTagId", 0)),
        NDRIntField("status", 0),
    ]


class RCreateServiceW_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRConfVarStrNullFieldUtf16("lpServiceName", ""),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpDisplayName", "")),
        NDRIntField("dwDesiredAccess", 0),
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwStartType", 0),
        NDRIntField("dwErrorControl", 0),
        NDRConfVarStrNullFieldUtf16("lpBinaryPathName", ""),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpLoadOrderGroup", "")),
        NDRFullPointerField(NDRIntField("lpdwTagId", 0)),
        NDRFullPointerField(
            NDRConfStrLenField(
                "lpDependencies", "", size_is=lambda pkt: pkt.dwDependSize
            )
        ),
        NDRIntField("dwDependSize", None, size_of="lpDependencies"),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpServiceStartName", "")),
        NDRFullPointerField(
            NDRConfStrLenField("lpPassword", "", size_is=lambda pkt: pkt.dwPwSize)
        ),
        NDRIntField("dwPwSize", None, size_of="lpPassword"),
    ]


class RCreateServiceW_Response(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRIntField("lpdwTagId", 0)),
        NDRPacketField("lpServiceHandle", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class REnumDependentServicesW_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwServiceState", 0),
        NDRIntField("cbBufSize", 0),
    ]


class REnumDependentServicesW_Response(NDRPacket):
    fields_desc = [
        NDRConfStrLenField("lpServices", "", size_is=lambda pkt: pkt.cbBufSize),
        NDRIntField("pcbBytesNeeded", 0),
        NDRIntField("lpServicesReturned", 0),
        NDRIntField("status", 0),
    ]


class REnumServicesStatusW_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwServiceState", 0),
        NDRIntField("cbBufSize", 0),
        NDRFullPointerField(NDRIntField("lpResumeIndex", 0)),
    ]


class REnumServicesStatusW_Response(NDRPacket):
    fields_desc = [
        NDRConfStrLenField("lpBuffer", "", size_is=lambda pkt: pkt.cbBufSize),
        NDRIntField("pcbBytesNeeded", 0),
        NDRIntField("lpServicesReturned", 0),
        NDRFullPointerField(NDRIntField("lpResumeIndex", 0)),
        NDRIntField("status", 0),
    ]


class ROpenSCManagerW_Request(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpMachineName", "")),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpDatabaseName", "")),
        NDRIntField("dwDesiredAccess", 0),
    ]


class ROpenSCManagerW_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("lpScHandle", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class ROpenServiceW_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRConfVarStrNullFieldUtf16("lpServiceName", ""),
        NDRIntField("dwDesiredAccess", 0),
    ]


class ROpenServiceW_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("lpServiceHandle", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class LPQUERY_SERVICE_CONFIGW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwStartType", 0),
        NDRIntField("dwErrorControl", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpBinaryPathName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpLoadOrderGroup", ""), deferred=True
        ),
        NDRIntField("dwTagId", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpDependencies", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpServiceStartName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpDisplayName", ""), deferred=True
        ),
    ]


class RQueryServiceConfigW_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("cbBufSize", 0),
    ]


class RQueryServiceConfigW_Response(NDRPacket):
    fields_desc = [
        NDRPacketField(
            "lpServiceConfig", LPQUERY_SERVICE_CONFIGW(), LPQUERY_SERVICE_CONFIGW
        ),
        NDRIntField("pcbBytesNeeded", 0),
        NDRIntField("status", 0),
    ]


class LPQUERY_SERVICE_LOCK_STATUSW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("fIsLocked", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpLockOwner", ""), deferred=True
        ),
        NDRIntField("dwLockDuration", 0),
    ]


class RQueryServiceLockStatusW_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRIntField("cbBufSize", 0),
    ]


class RQueryServiceLockStatusW_Response(NDRPacket):
    fields_desc = [
        NDRPacketField(
            "lpLockStatus", LPQUERY_SERVICE_LOCK_STATUSW(), LPQUERY_SERVICE_LOCK_STATUSW
        ),
        NDRIntField("pcbBytesNeeded", 0),
        NDRIntField("status", 0),
    ]


class LPSTRING_PTRSW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("StringPtr", ""), deferred=True)
    ]


class RStartServiceW_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("argc", None, size_of="argv"),
        NDRFullPointerField(
            NDRConfPacketListField(
                "argv", [LPSTRING_PTRSW()], LPSTRING_PTRSW, size_is=lambda pkt: pkt.argc
            )
        ),
    ]


class RStartServiceW_Response(NDRPacket):
    fields_desc = [NDRIntField("status", 0)]


class RGetServiceDisplayNameW_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRConfVarStrNullFieldUtf16("lpServiceName", ""),
        NDRIntField("lpcchBuffer", 0),
    ]


class RGetServiceDisplayNameW_Response(NDRPacket):
    fields_desc = [
        NDRConfVarStrLenFieldUtf16(
            "lpDisplayName", "", size_is=lambda pkt: (pkt.lpcchBuffer + 1)
        ),
        NDRIntField(
            "lpcchBuffer", None, size_of="lpDisplayName", adjust=lambda _, x: (x - 1)
        ),
        NDRIntField("status", 0),
    ]


class RGetServiceKeyNameW_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRConfVarStrNullFieldUtf16("lpDisplayName", ""),
        NDRIntField("lpcchBuffer", 0),
    ]


class RGetServiceKeyNameW_Response(NDRPacket):
    fields_desc = [
        NDRConfVarStrLenFieldUtf16(
            "lpServiceName", "", size_is=lambda pkt: (pkt.lpcchBuffer + 1)
        ),
        NDRIntField(
            "lpcchBuffer", None, size_of="lpServiceName", adjust=lambda _, x: (x - 1)
        ),
        NDRIntField("status", 0),
    ]


class RChangeServiceConfigA_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwStartType", 0),
        NDRIntField("dwErrorControl", 0),
        NDRFullPointerField(NDRConfVarStrNullField("lpBinaryPathName", "")),
        NDRFullPointerField(NDRConfVarStrNullField("lpLoadOrderGroup", "")),
        NDRFullPointerField(NDRIntField("lpdwTagId", 0)),
        NDRFullPointerField(
            NDRConfStrLenField(
                "lpDependencies", "", size_is=lambda pkt: pkt.dwDependSize
            )
        ),
        NDRIntField("dwDependSize", None, size_of="lpDependencies"),
        NDRFullPointerField(NDRConfVarStrNullField("lpServiceStartName", "")),
        NDRFullPointerField(
            NDRConfStrLenField("lpPassword", "", size_is=lambda pkt: pkt.dwPwSize)
        ),
        NDRIntField("dwPwSize", None, size_of="lpPassword"),
        NDRFullPointerField(NDRConfVarStrNullField("lpDisplayName", "")),
    ]


class RChangeServiceConfigA_Response(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRIntField("lpdwTagId", 0)),
        NDRIntField("status", 0),
    ]


class RCreateServiceA_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRConfVarStrNullField("lpServiceName", ""),
        NDRFullPointerField(NDRConfVarStrNullField("lpDisplayName", "")),
        NDRIntField("dwDesiredAccess", 0),
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwStartType", 0),
        NDRIntField("dwErrorControl", 0),
        NDRConfVarStrNullField("lpBinaryPathName", ""),
        NDRFullPointerField(NDRConfVarStrNullField("lpLoadOrderGroup", "")),
        NDRFullPointerField(NDRIntField("lpdwTagId", 0)),
        NDRFullPointerField(
            NDRConfStrLenField(
                "lpDependencies", "", size_is=lambda pkt: pkt.dwDependSize
            )
        ),
        NDRIntField("dwDependSize", None, size_of="lpDependencies"),
        NDRFullPointerField(NDRConfVarStrNullField("lpServiceStartName", "")),
        NDRFullPointerField(
            NDRConfStrLenField("lpPassword", "", size_is=lambda pkt: pkt.dwPwSize)
        ),
        NDRIntField("dwPwSize", None, size_of="lpPassword"),
    ]


class RCreateServiceA_Response(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRIntField("lpdwTagId", 0)),
        NDRPacketField("lpServiceHandle", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class REnumDependentServicesA_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwServiceState", 0),
        NDRIntField("cbBufSize", 0),
    ]


class REnumDependentServicesA_Response(NDRPacket):
    fields_desc = [
        NDRConfStrLenField("lpServices", "", size_is=lambda pkt: pkt.cbBufSize),
        NDRIntField("pcbBytesNeeded", 0),
        NDRIntField("lpServicesReturned", 0),
        NDRIntField("status", 0),
    ]


class REnumServicesStatusA_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwServiceState", 0),
        NDRIntField("cbBufSize", 0),
        NDRFullPointerField(NDRIntField("lpResumeIndex", 0)),
    ]


class REnumServicesStatusA_Response(NDRPacket):
    fields_desc = [
        NDRConfStrLenField("lpBuffer", "", size_is=lambda pkt: pkt.cbBufSize),
        NDRIntField("pcbBytesNeeded", 0),
        NDRIntField("lpServicesReturned", 0),
        NDRFullPointerField(NDRIntField("lpResumeIndex", 0)),
        NDRIntField("status", 0),
    ]


class ROpenSCManagerA_Request(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullField("lpMachineName", "")),
        NDRFullPointerField(NDRConfVarStrNullField("lpDatabaseName", "")),
        NDRIntField("dwDesiredAccess", 0),
    ]


class ROpenSCManagerA_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("lpScHandle", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class ROpenServiceA_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRConfVarStrNullField("lpServiceName", ""),
        NDRIntField("dwDesiredAccess", 0),
    ]


class ROpenServiceA_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("lpServiceHandle", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class LPQUERY_SERVICE_CONFIGA(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwStartType", 0),
        NDRIntField("dwErrorControl", 0),
        NDRFullPointerField(
            NDRConfVarStrNullField("lpBinaryPathName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullField("lpLoadOrderGroup", ""), deferred=True
        ),
        NDRIntField("dwTagId", 0),
        NDRFullPointerField(
            NDRConfVarStrNullField("lpDependencies", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullField("lpServiceStartName", ""), deferred=True
        ),
        NDRFullPointerField(NDRConfVarStrNullField("lpDisplayName", ""), deferred=True),
    ]


class RQueryServiceConfigA_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("cbBufSize", 0),
    ]


class RQueryServiceConfigA_Response(NDRPacket):
    fields_desc = [
        NDRPacketField(
            "lpServiceConfig", LPQUERY_SERVICE_CONFIGA(), LPQUERY_SERVICE_CONFIGA
        ),
        NDRIntField("pcbBytesNeeded", 0),
        NDRIntField("status", 0),
    ]


class LPQUERY_SERVICE_LOCK_STATUSA(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("fIsLocked", 0),
        NDRFullPointerField(NDRConfVarStrNullField("lpLockOwner", ""), deferred=True),
        NDRIntField("dwLockDuration", 0),
    ]


class RQueryServiceLockStatusA_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRIntField("cbBufSize", 0),
    ]


class RQueryServiceLockStatusA_Response(NDRPacket):
    fields_desc = [
        NDRPacketField(
            "lpLockStatus", LPQUERY_SERVICE_LOCK_STATUSA(), LPQUERY_SERVICE_LOCK_STATUSA
        ),
        NDRIntField("pcbBytesNeeded", 0),
        NDRIntField("status", 0),
    ]


class LPSTRING_PTRSA(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullField("StringPtr", ""), deferred=True)
    ]


class RStartServiceA_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("argc", None, size_of="argv"),
        NDRFullPointerField(
            NDRConfPacketListField(
                "argv", [LPSTRING_PTRSA()], LPSTRING_PTRSA, size_is=lambda pkt: pkt.argc
            )
        ),
    ]


class RStartServiceA_Response(NDRPacket):
    fields_desc = [NDRIntField("status", 0)]


class RGetServiceDisplayNameA_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRConfVarStrNullField("lpServiceName", ""),
        NDRIntField("lpcchBuffer", 0),
    ]


class RGetServiceDisplayNameA_Response(NDRPacket):
    fields_desc = [
        NDRConfVarStrNullField("lpDisplayName", ""),
        NDRIntField("lpcchBuffer", None, size_of="lpDisplayName"),
        NDRIntField("status", 0),
    ]


class RGetServiceKeyNameA_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRConfVarStrNullField("lpDisplayName", ""),
        NDRIntField("lpcchBuffer", 0),
    ]


class RGetServiceKeyNameA_Response(NDRPacket):
    fields_desc = [
        NDRConfVarStrNullField("lpKeyName", ""),
        NDRIntField("lpcchBuffer", None, size_of="lpKeyName"),
        NDRIntField("status", 0),
    ]


class REnumServiceGroupW_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwServiceState", 0),
        NDRIntField("cbBufSize", 0),
        NDRFullPointerField(NDRIntField("lpResumeIndex", 0)),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("pszGroupName", "")),
    ]


class REnumServiceGroupW_Response(NDRPacket):
    fields_desc = [
        NDRConfStrLenField("lpBuffer", "", size_is=lambda pkt: pkt.cbBufSize),
        NDRIntField("pcbBytesNeeded", 0),
        NDRIntField("lpServicesReturned", 0),
        NDRFullPointerField(NDRIntField("lpResumeIndex", 0)),
        NDRIntField("status", 0),
    ]


class LPSERVICE_DESCRIPTIONA(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullField("lpDescription", ""), deferred=True)
    ]


class SC_ACTION_TYPE(IntEnum):
    SC_ACTION_NONE = 0
    SC_ACTION_RESTART = 1
    SC_ACTION_REBOOT = 2
    SC_ACTION_RUN_COMMAND = 3


class SC_ACTION(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [NDRIntEnumField("Type", 0, SC_ACTION_TYPE), NDRIntField("Delay", 0)]


class LPSERVICE_FAILURE_ACTIONSA(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwResetPeriod", 0),
        NDRFullPointerField(NDRConfVarStrNullField("lpRebootMsg", ""), deferred=True),
        NDRFullPointerField(NDRConfVarStrNullField("lpCommand", ""), deferred=True),
        NDRIntField("cActions", None, size_of="lpsaActions"),
        NDRFullPointerField(
            NDRConfPacketListField(
                "lpsaActions", [], SC_ACTION, size_is=lambda pkt: pkt.cActions
            ),
            deferred=True,
        ),
    ]


class LPSERVICE_DELAYED_AUTO_START_INFO(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [NDRSignedIntField("fDelayedAutostart", 0)]


class LPSERVICE_FAILURE_ACTIONS_FLAG(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [NDRSignedIntField("fFailureActionsOnNonCrashFailures", 0)]


class LPSERVICE_SID_INFO(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [NDRIntField("dwServiceSidType", 0)]


class LPSERVICE_RPC_REQUIRED_PRIVILEGES_INFO(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("cbRequiredPrivileges", None, size_of="pRequiredPrivileges"),
        NDRFullPointerField(
            NDRConfStrLenField(
                "pRequiredPrivileges", "", size_is=lambda pkt: pkt.cbRequiredPrivileges
            ),
            deferred=True,
        ),
    ]


class LPSERVICE_PRESHUTDOWN_INFO(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [NDRIntField("dwPreshutdownTimeout", 0)]


class GUID(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRIntField("Data1", 0),
        NDRShortField("Data2", 0),
        NDRShortField("Data3", 0),
        StrFixedLenField("Data4", "", length=8),
    ]


class PSERVICE_TRIGGER_SPECIFIC_DATA_ITEM(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwDataType", 0),
        NDRIntField("cbData", None, size_of="pData"),
        NDRFullPointerField(
            NDRConfStrLenField("pData", "", size_is=lambda pkt: pkt.cbData),
            deferred=True,
        ),
    ]


class PSERVICE_TRIGGER(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwTriggerType", 0),
        NDRIntField("dwAction", 0),
        NDRFullPointerField(
            NDRPacketField("pTriggerSubtype", GUID(), GUID), deferred=True
        ),
        NDRIntField("cDataItems", None, size_of="pDataItems"),
        NDRFullPointerField(
            NDRConfPacketListField(
                "pDataItems",
                [PSERVICE_TRIGGER_SPECIFIC_DATA_ITEM()],
                PSERVICE_TRIGGER_SPECIFIC_DATA_ITEM,
                size_is=lambda pkt: pkt.cDataItems,
            ),
            deferred=True,
        ),
    ]


class PSERVICE_TRIGGER_INFO(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("cTriggers", None, size_of="pTriggers"),
        NDRFullPointerField(
            NDRConfPacketListField(
                "pTriggers",
                [PSERVICE_TRIGGER()],
                PSERVICE_TRIGGER,
                size_is=lambda pkt: pkt.cTriggers,
            ),
            deferred=True,
        ),
        NDRFullPointerField(NDRByteField("pReserved", 0), deferred=True),
    ]


class LPSERVICE_PREFERRED_NODE_INFO(NDRPacket):
    ALIGNMENT = (2, 2)
    fields_desc = [NDRShortField("usPreferredNode", 0), NDRByteField("fDelete", 0)]


class SC_RPC_CONFIG_INFOA(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwInfoLevel", 0),
        NDRUnionField(
            [
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value", LPSERVICE_DESCRIPTIONA(), LPSERVICE_DESCRIPTIONA
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 1),
                        (lambda _, val: val.tag == 1),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value",
                            LPSERVICE_FAILURE_ACTIONSA(),
                            LPSERVICE_FAILURE_ACTIONSA,
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 2),
                        (lambda _, val: val.tag == 2),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value",
                            LPSERVICE_DELAYED_AUTO_START_INFO(),
                            LPSERVICE_DELAYED_AUTO_START_INFO,
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 3),
                        (lambda _, val: val.tag == 3),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value",
                            LPSERVICE_FAILURE_ACTIONS_FLAG(),
                            LPSERVICE_FAILURE_ACTIONS_FLAG,
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 4),
                        (lambda _, val: val.tag == 4),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value", LPSERVICE_SID_INFO(), LPSERVICE_SID_INFO
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 5),
                        (lambda _, val: val.tag == 5),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value",
                            LPSERVICE_RPC_REQUIRED_PRIVILEGES_INFO(),
                            LPSERVICE_RPC_REQUIRED_PRIVILEGES_INFO,
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 6),
                        (lambda _, val: val.tag == 6),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value",
                            LPSERVICE_PRESHUTDOWN_INFO(),
                            LPSERVICE_PRESHUTDOWN_INFO,
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 7),
                        (lambda _, val: val.tag == 7),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value", PSERVICE_TRIGGER_INFO(), PSERVICE_TRIGGER_INFO
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 8),
                        (lambda _, val: val.tag == 8),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value",
                            LPSERVICE_PREFERRED_NODE_INFO(),
                            LPSERVICE_PREFERRED_NODE_INFO,
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 9),
                        (lambda _, val: val.tag == 9),
                    ),
                ),
            ],
            StrFixedLenField("value", "", length=0),
            align=(4, 8),
            switch_fmt=("L", "L"),
        ),
    ]


class RChangeServiceConfig2A_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRPacketField("Info", SC_RPC_CONFIG_INFOA(), SC_RPC_CONFIG_INFOA),
    ]


class RChangeServiceConfig2A_Response(NDRPacket):
    fields_desc = [NDRIntField("status", 0)]


class LPSERVICE_DESCRIPTIONW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpDescription", ""), deferred=True
        )
    ]


class LPSERVICE_FAILURE_ACTIONSW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwResetPeriod", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpRebootMsg", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpCommand", ""), deferred=True
        ),
        NDRIntField("cActions", None, size_of="lpsaActions"),
        NDRFullPointerField(
            NDRConfPacketListField(
                "lpsaActions", [], SC_ACTION, size_is=lambda pkt: pkt.cActions
            ),
            deferred=True,
        ),
    ]


class SC_RPC_CONFIG_INFOW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwInfoLevel", 0),
        NDRUnionField(
            [
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value", LPSERVICE_DESCRIPTIONW(), LPSERVICE_DESCRIPTIONW
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 1),
                        (lambda _, val: val.tag == 1),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value",
                            LPSERVICE_FAILURE_ACTIONSW(),
                            LPSERVICE_FAILURE_ACTIONSW,
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 2),
                        (lambda _, val: val.tag == 2),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value",
                            LPSERVICE_DELAYED_AUTO_START_INFO(),
                            LPSERVICE_DELAYED_AUTO_START_INFO,
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 3),
                        (lambda _, val: val.tag == 3),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value",
                            LPSERVICE_FAILURE_ACTIONS_FLAG(),
                            LPSERVICE_FAILURE_ACTIONS_FLAG,
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 4),
                        (lambda _, val: val.tag == 4),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value", LPSERVICE_SID_INFO(), LPSERVICE_SID_INFO
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 5),
                        (lambda _, val: val.tag == 5),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value",
                            LPSERVICE_RPC_REQUIRED_PRIVILEGES_INFO(),
                            LPSERVICE_RPC_REQUIRED_PRIVILEGES_INFO,
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 6),
                        (lambda _, val: val.tag == 6),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value",
                            LPSERVICE_PRESHUTDOWN_INFO(),
                            LPSERVICE_PRESHUTDOWN_INFO,
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 7),
                        (lambda _, val: val.tag == 7),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value", PSERVICE_TRIGGER_INFO(), PSERVICE_TRIGGER_INFO
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 8),
                        (lambda _, val: val.tag == 8),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value",
                            LPSERVICE_PREFERRED_NODE_INFO(),
                            LPSERVICE_PREFERRED_NODE_INFO,
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 9),
                        (lambda _, val: val.tag == 9),
                    ),
                ),
            ],
            StrFixedLenField("value", "", length=0),
            align=(4, 8),
            switch_fmt=("L", "L"),
        ),
    ]


class RChangeServiceConfig2W_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRPacketField("Info", SC_RPC_CONFIG_INFOW(), SC_RPC_CONFIG_INFOW),
    ]


class RChangeServiceConfig2W_Response(NDRPacket):
    fields_desc = [NDRIntField("status", 0)]


class RQueryServiceConfig2A_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwInfoLevel", 0),
        NDRIntField("cbBufSize", 0),
    ]


class RQueryServiceConfig2A_Response(NDRPacket):
    fields_desc = [
        NDRConfStrLenField("lpBuffer", "", size_is=lambda pkt: pkt.cbBufSize),
        NDRIntField("pcbBytesNeeded", 0),
        NDRIntField("status", 0),
    ]


class RQueryServiceConfig2W_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwInfoLevel", 0),
        NDRIntField("cbBufSize", 0),
    ]


class RQueryServiceConfig2W_Response(NDRPacket):
    fields_desc = [
        NDRConfStrLenField("lpBuffer", "", size_is=lambda pkt: pkt.cbBufSize),
        NDRIntField("pcbBytesNeeded", 0),
        NDRIntField("status", 0),
    ]


class SC_STATUS_TYPE(IntEnum):
    SC_STATUS_PROCESS_INFO = 0


class RQueryServiceStatusEx_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntEnumField("InfoLevel", 0, SC_STATUS_TYPE),
        NDRIntField("cbBufSize", 0),
    ]


class RQueryServiceStatusEx_Response(NDRPacket):
    fields_desc = [
        NDRConfStrLenField("lpBuffer", "", size_is=lambda pkt: pkt.cbBufSize),
        NDRIntField("pcbBytesNeeded", 0),
        NDRIntField("status", 0),
    ]


class SC_ENUM_TYPE(IntEnum):
    SC_ENUM_PROCESS_INFO = 0


class REnumServicesStatusExA_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRIntEnumField("InfoLevel", 0, SC_ENUM_TYPE),
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwServiceState", 0),
        NDRIntField("cbBufSize", 0),
        NDRFullPointerField(NDRIntField("lpResumeIndex", 0)),
        NDRFullPointerField(NDRConfVarStrNullField("pszGroupName", "")),
    ]


class REnumServicesStatusExA_Response(NDRPacket):
    fields_desc = [
        NDRConfStrLenField("lpBuffer", "", size_is=lambda pkt: pkt.cbBufSize),
        NDRIntField("pcbBytesNeeded", 0),
        NDRIntField("lpServicesReturned", 0),
        NDRFullPointerField(NDRIntField("lpResumeIndex", 0)),
        NDRIntField("status", 0),
    ]


class REnumServicesStatusExW_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRIntEnumField("InfoLevel", 0, SC_ENUM_TYPE),
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwServiceState", 0),
        NDRIntField("cbBufSize", 0),
        NDRFullPointerField(NDRIntField("lpResumeIndex", 0)),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("pszGroupName", "")),
    ]


class REnumServicesStatusExW_Response(NDRPacket):
    fields_desc = [
        NDRConfStrLenField("lpBuffer", "", size_is=lambda pkt: pkt.cbBufSize),
        NDRIntField("pcbBytesNeeded", 0),
        NDRIntField("lpServicesReturned", 0),
        NDRFullPointerField(NDRIntField("lpResumeIndex", 0)),
        NDRIntField("status", 0),
    ]


class RCreateServiceWOW64A_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRConfVarStrNullField("lpServiceName", ""),
        NDRFullPointerField(NDRConfVarStrNullField("lpDisplayName", "")),
        NDRIntField("dwDesiredAccess", 0),
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwStartType", 0),
        NDRIntField("dwErrorControl", 0),
        NDRConfVarStrNullField("lpBinaryPathName", ""),
        NDRFullPointerField(NDRConfVarStrNullField("lpLoadOrderGroup", "")),
        NDRFullPointerField(NDRIntField("lpdwTagId", 0)),
        NDRFullPointerField(
            NDRConfStrLenField(
                "lpDependencies", "", size_is=lambda pkt: pkt.dwDependSize
            )
        ),
        NDRIntField("dwDependSize", None, size_of="lpDependencies"),
        NDRFullPointerField(NDRConfVarStrNullField("lpServiceStartName", "")),
        NDRFullPointerField(
            NDRConfStrLenField("lpPassword", "", size_is=lambda pkt: pkt.dwPwSize)
        ),
        NDRIntField("dwPwSize", None, size_of="lpPassword"),
    ]


class RCreateServiceWOW64A_Response(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRIntField("lpdwTagId", 0)),
        NDRPacketField("lpServiceHandle", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class RCreateServiceWOW64W_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRConfVarStrNullFieldUtf16("lpServiceName", ""),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpDisplayName", "")),
        NDRIntField("dwDesiredAccess", 0),
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwStartType", 0),
        NDRIntField("dwErrorControl", 0),
        NDRConfVarStrNullFieldUtf16("lpBinaryPathName", ""),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpLoadOrderGroup", "")),
        NDRFullPointerField(NDRIntField("lpdwTagId", 0)),
        NDRFullPointerField(
            NDRConfStrLenField(
                "lpDependencies", "", size_is=lambda pkt: pkt.dwDependSize
            )
        ),
        NDRIntField("dwDependSize", None, size_of="lpDependencies"),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpServiceStartName", "")),
        NDRFullPointerField(
            NDRConfStrLenField("lpPassword", "", size_is=lambda pkt: pkt.dwPwSize)
        ),
        NDRIntField("dwPwSize", None, size_of="lpPassword"),
    ]


class RCreateServiceWOW64W_Response(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRIntField("lpdwTagId", 0)),
        NDRPacketField("lpServiceHandle", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class SERVICE_STATUS_PROCESS(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwCurrentState", 0),
        NDRIntField("dwControlsAccepted", 0),
        NDRIntField("dwWin32ExitCode", 0),
        NDRIntField("dwServiceSpecificExitCode", 0),
        NDRIntField("dwCheckPoint", 0),
        NDRIntField("dwWaitHint", 0),
        NDRIntField("dwProcessId", 0),
        NDRIntField("dwServiceFlags", 0),
    ]


class PSERVICE_NOTIFY_STATUS_CHANGE_PARAMS_1(NDRPacket):
    ALIGNMENT = (8, 8)
    fields_desc = [
        NDRLongField("ullThreadId", 0),
        NDRIntField("dwNotifyMask", 0),
        StrFixedLenField("CallbackAddressArray", "", length=16),
        StrFixedLenField("CallbackParamAddressArray", "", length=16),
        NDRPacketField(
            "ServiceStatus", SERVICE_STATUS_PROCESS(), SERVICE_STATUS_PROCESS
        ),
        NDRIntField("dwNotificationStatus", 0),
        NDRIntField("dwSequence", 0),
    ]


class PSERVICE_NOTIFY_STATUS_CHANGE_PARAMS_2(NDRPacket):
    ALIGNMENT = (8, 8)
    fields_desc = [
        NDRLongField("ullThreadId", 0),
        NDRIntField("dwNotifyMask", 0),
        StrFixedLenField("CallbackAddressArray", "", length=16),
        StrFixedLenField("CallbackParamAddressArray", "", length=16),
        NDRPacketField(
            "ServiceStatus", SERVICE_STATUS_PROCESS(), SERVICE_STATUS_PROCESS
        ),
        NDRIntField("dwNotificationStatus", 0),
        NDRIntField("dwSequence", 0),
        NDRIntField("dwNotificationTriggered", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("pszServiceNames", ""), deferred=True
        ),
    ]


class SC_RPC_NOTIFY_PARAMS(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwInfoLevel", 0),
        NDRUnionField(
            [
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value",
                            PSERVICE_NOTIFY_STATUS_CHANGE_PARAMS_1(),
                            PSERVICE_NOTIFY_STATUS_CHANGE_PARAMS_1,
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 1),
                        (lambda _, val: val.tag == 1),
                    ),
                ),
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "value",
                            PSERVICE_NOTIFY_STATUS_CHANGE_PARAMS_2(),
                            PSERVICE_NOTIFY_STATUS_CHANGE_PARAMS_2,
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 2),
                        (lambda _, val: val.tag == 2),
                    ),
                ),
            ],
            StrFixedLenField("value", "", length=0),
            align=(4, 8),
            switch_fmt=("L", "L"),
        ),
    ]


class RNotifyServiceStatusChange_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRPacketField("NotifyParams", SC_RPC_NOTIFY_PARAMS(), SC_RPC_NOTIFY_PARAMS),
        NDRPacketField("pClientProcessGuid", GUID(), GUID),
    ]


class RNotifyServiceStatusChange_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("pSCMProcessGuid", GUID(), GUID),
        NDRSignedIntField("pfCreateRemoteQueue", 0),
        NDRPacketField("phNotify", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class PSC_RPC_NOTIFY_PARAMS_LIST(NDRPacket):
    ALIGNMENT = (4, 8)
    DEPORTED_CONFORMANTS = ["NotifyParamsArray"]
    fields_desc = [
        NDRIntField("cElements", None, size_of="NotifyParamsArray"),
        NDRConfPacketListField(
            "NotifyParamsArray",
            [],
            SC_RPC_NOTIFY_PARAMS,
            size_is=lambda pkt: pkt.cElements,
            conformant_in_struct=True,
        ),
    ]


class RGetNotifyResults_Request(NDRPacket):
    fields_desc = [NDRPacketField("hNotify", NDRContextHandle(), NDRContextHandle)]


class RGetNotifyResults_Response(NDRPacket):
    fields_desc = [
        NDRFullPointerField(
            NDRPacketField(
                "ppNotifyParams",
                PSC_RPC_NOTIFY_PARAMS_LIST(),
                PSC_RPC_NOTIFY_PARAMS_LIST,
            )
        ),
        NDRIntField("status", 0),
    ]


class RCloseNotifyHandle_Request(NDRPacket):
    fields_desc = [NDRPacketField("phNotify", NDRContextHandle(), NDRContextHandle)]


class RCloseNotifyHandle_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("phNotify", NDRContextHandle(), NDRContextHandle),
        NDRSignedIntField("pfApcFired", 0),
        NDRIntField("status", 0),
    ]


class PSERVICE_CONTROL_STATUS_REASON_IN_PARAMSA(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwReason", 0),
        NDRFullPointerField(NDRConfVarStrNullField("pszComment", ""), deferred=True),
    ]


class PSERVICE_CONTROL_STATUS_REASON_OUT_PARAMS(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRPacketField(
            "ServiceStatus", SERVICE_STATUS_PROCESS(), SERVICE_STATUS_PROCESS
        )
    ]


class RControlServiceExA_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwControl", 0),
        NDRIntField("dwInfoLevel", 0),
        NDRUnionField(
            [
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "pControlInParams",
                            PSERVICE_CONTROL_STATUS_REASON_IN_PARAMSA(),
                            PSERVICE_CONTROL_STATUS_REASON_IN_PARAMSA,
                        )
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 1),
                        (lambda _, val: val.tag == 1),
                    ),
                )
            ],
            StrFixedLenField("pControlInParams", "", length=0),
            align=(4, 8),
            switch_fmt=("L", "L"),
        ),
    ]


class RControlServiceExA_Response(NDRPacket):
    fields_desc = [
        NDRUnionField(
            [
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "pControlOutParams",
                            PSERVICE_CONTROL_STATUS_REASON_OUT_PARAMS(),
                            PSERVICE_CONTROL_STATUS_REASON_OUT_PARAMS,
                        )
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 1),
                        (lambda _, val: val.tag == 1),
                    ),
                )
            ],
            StrFixedLenField("pControlOutParams", "", length=0),
            align=(4, 8),
            switch_fmt=("L", "L"),
        ),
        NDRIntField("status", 0),
    ]


class PSERVICE_CONTROL_STATUS_REASON_IN_PARAMSW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwReason", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("pszComment", ""), deferred=True
        ),
    ]


class RControlServiceExW_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwControl", 0),
        NDRIntField("dwInfoLevel", 0),
        NDRUnionField(
            [
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "pControlInParams",
                            PSERVICE_CONTROL_STATUS_REASON_IN_PARAMSW(),
                            PSERVICE_CONTROL_STATUS_REASON_IN_PARAMSW,
                        )
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 1),
                        (lambda _, val: val.tag == 1),
                    ),
                )
            ],
            StrFixedLenField("pControlInParams", "", length=0),
            align=(4, 8),
            switch_fmt=("L", "L"),
        ),
    ]


class RControlServiceExW_Response(NDRPacket):
    fields_desc = [
        NDRUnionField(
            [
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "pControlOutParams",
                            PSERVICE_CONTROL_STATUS_REASON_OUT_PARAMS(),
                            PSERVICE_CONTROL_STATUS_REASON_OUT_PARAMS,
                        )
                    ),
                    (
                        (lambda pkt: getattr(pkt, "dwInfoLevel", None) == 1),
                        (lambda _, val: val.tag == 1),
                    ),
                )
            ],
            StrFixedLenField("pControlOutParams", "", length=0),
            align=(4, 8),
            switch_fmt=("L", "L"),
        ),
        NDRIntField("status", 0),
    ]


class RQueryServiceConfigEx_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hService", NDRContextHandle(), NDRContextHandle),
        NDRIntField("dwInfoLevel", 0),
    ]


class RQueryServiceConfigEx_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("pInfo", SC_RPC_CONFIG_INFOW(), SC_RPC_CONFIG_INFOW),
        NDRIntField("status", 0),
    ]


class RCreateWowService_Request(NDRPacket):
    fields_desc = [
        NDRPacketField("hSCManager", NDRContextHandle(), NDRContextHandle),
        NDRConfVarStrNullFieldUtf16("lpServiceName", ""),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpDisplayName", "")),
        NDRIntField("dwDesiredAccess", 0),
        NDRIntField("dwServiceType", 0),
        NDRIntField("dwStartType", 0),
        NDRIntField("dwErrorControl", 0),
        NDRConfVarStrNullFieldUtf16("lpBinaryPathName", ""),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpLoadOrderGroup", "")),
        NDRFullPointerField(NDRIntField("lpdwTagId", 0)),
        NDRFullPointerField(
            NDRConfStrLenField(
                "lpDependencies", "", size_is=lambda pkt: pkt.dwDependSize
            )
        ),
        NDRIntField("dwDependSize", None, size_of="lpDependencies"),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("lpServiceStartName", "")),
        NDRFullPointerField(
            NDRConfStrLenField("lpPassword", "", size_is=lambda pkt: pkt.dwPwSize)
        ),
        NDRIntField("dwPwSize", None, size_of="lpPassword"),
        NDRShortField("dwServiceWowType", 0),
    ]


class RCreateWowService_Response(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRIntField("lpdwTagId", 0)),
        NDRPacketField("lpServiceHandle", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class ROpenSCManager2_Request(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("DatabaseName", "")),
        NDRIntField("DesiredAccess", 0),
    ]


class ROpenSCManager2_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("ScmHandle", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


SVCCTL_OPNUMS = {
    0: DceRpcOp(RCloseServiceHandle_Request, RCloseServiceHandle_Response),
    1: DceRpcOp(RControlService_Request, RControlService_Response),
    2: DceRpcOp(RDeleteService_Request, RDeleteService_Response),
    3: DceRpcOp(RLockServiceDatabase_Request, RLockServiceDatabase_Response),
    4: DceRpcOp(
        RQueryServiceObjectSecurity_Request, RQueryServiceObjectSecurity_Response
    ),
    5: DceRpcOp(RSetServiceObjectSecurity_Request, RSetServiceObjectSecurity_Response),
    6: DceRpcOp(RQueryServiceStatus_Request, RQueryServiceStatus_Response),
    7: DceRpcOp(RSetServiceStatus_Request, RSetServiceStatus_Response),
    8: DceRpcOp(RUnlockServiceDatabase_Request, RUnlockServiceDatabase_Response),
    9: DceRpcOp(RNotifyBootConfigStatus_Request, RNotifyBootConfigStatus_Response),
    # 10: Opnum10NotUsedOnWire,
    11: DceRpcOp(RChangeServiceConfigW_Request, RChangeServiceConfigW_Response),
    12: DceRpcOp(RCreateServiceW_Request, RCreateServiceW_Response),
    13: DceRpcOp(REnumDependentServicesW_Request, REnumDependentServicesW_Response),
    14: DceRpcOp(REnumServicesStatusW_Request, REnumServicesStatusW_Response),
    15: DceRpcOp(ROpenSCManagerW_Request, ROpenSCManagerW_Response),
    16: DceRpcOp(ROpenServiceW_Request, ROpenServiceW_Response),
    17: DceRpcOp(RQueryServiceConfigW_Request, RQueryServiceConfigW_Response),
    18: DceRpcOp(RQueryServiceLockStatusW_Request, RQueryServiceLockStatusW_Response),
    19: DceRpcOp(RStartServiceW_Request, RStartServiceW_Response),
    20: DceRpcOp(RGetServiceDisplayNameW_Request, RGetServiceDisplayNameW_Response),
    21: DceRpcOp(RGetServiceKeyNameW_Request, RGetServiceKeyNameW_Response),
    # 22: Opnum22NotUsedOnWire,
    23: DceRpcOp(RChangeServiceConfigA_Request, RChangeServiceConfigA_Response),
    24: DceRpcOp(RCreateServiceA_Request, RCreateServiceA_Response),
    25: DceRpcOp(REnumDependentServicesA_Request, REnumDependentServicesA_Response),
    26: DceRpcOp(REnumServicesStatusA_Request, REnumServicesStatusA_Response),
    27: DceRpcOp(ROpenSCManagerA_Request, ROpenSCManagerA_Response),
    28: DceRpcOp(ROpenServiceA_Request, ROpenServiceA_Response),
    29: DceRpcOp(RQueryServiceConfigA_Request, RQueryServiceConfigA_Response),
    30: DceRpcOp(RQueryServiceLockStatusA_Request, RQueryServiceLockStatusA_Response),
    31: DceRpcOp(RStartServiceA_Request, RStartServiceA_Response),
    32: DceRpcOp(RGetServiceDisplayNameA_Request, RGetServiceDisplayNameA_Response),
    33: DceRpcOp(RGetServiceKeyNameA_Request, RGetServiceKeyNameA_Response),
    # 34: Opnum34NotUsedOnWire,
    35: DceRpcOp(REnumServiceGroupW_Request, REnumServiceGroupW_Response),
    36: DceRpcOp(RChangeServiceConfig2A_Request, RChangeServiceConfig2A_Response),
    37: DceRpcOp(RChangeServiceConfig2W_Request, RChangeServiceConfig2W_Response),
    38: DceRpcOp(RQueryServiceConfig2A_Request, RQueryServiceConfig2A_Response),
    39: DceRpcOp(RQueryServiceConfig2W_Request, RQueryServiceConfig2W_Response),
    40: DceRpcOp(RQueryServiceStatusEx_Request, RQueryServiceStatusEx_Response),
    41: DceRpcOp(REnumServicesStatusExA_Request, REnumServicesStatusExA_Response),
    42: DceRpcOp(REnumServicesStatusExW_Request, REnumServicesStatusExW_Response),
    # 43: Opnum43NotUsedOnWire,
    44: DceRpcOp(RCreateServiceWOW64A_Request, RCreateServiceWOW64A_Response),
    45: DceRpcOp(RCreateServiceWOW64W_Request, RCreateServiceWOW64W_Response),
    # 46: Opnum46NotUsedOnWire,
    47: DceRpcOp(
        RNotifyServiceStatusChange_Request, RNotifyServiceStatusChange_Response
    ),
    48: DceRpcOp(RGetNotifyResults_Request, RGetNotifyResults_Response),
    49: DceRpcOp(RCloseNotifyHandle_Request, RCloseNotifyHandle_Response),
    50: DceRpcOp(RControlServiceExA_Request, RControlServiceExA_Response),
    51: DceRpcOp(RControlServiceExW_Request, RControlServiceExW_Response),
    # 52: Opnum52NotUsedOnWire,
    # 53: Opnum53NotUsedOnWire,
    # 54: Opnum54NotUsedOnWire,
    # 55: Opnum55NotUsedOnWire,
    56: DceRpcOp(RQueryServiceConfigEx_Request, RQueryServiceConfigEx_Response),
    # 57: Opnum57NotUsedOnWire,
    # 58: Opnum58NotUsedOnWire,
    # 59: Opnum59NotUsedOnWire,
    60: DceRpcOp(RCreateWowService_Request, RCreateWowService_Response),
    # 61: Opnum61NotUsedOnWire,
    # 62: Opnum62NotUsedOnWire,
    # 63: Opnum63NotUsedOnWire,
    64: DceRpcOp(ROpenSCManager2_Request, ROpenSCManager2_Response),
}
register_dcerpc_interface(
    name="svcctl",
    uuid=uuid.UUID("367ABB81-9844-35F1-AD32-98F038001003"),
    version="2.0",
    opnums=SVCCTL_OPNUMS,
)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
[MS-EERR] ExtendedError Remote Data Structure
"""

# Wireshark does not know how to read this !

import uuid
from enum import IntEnum

from scapy.fields import StrFixedLenField, UTCTimeField
from scapy.layers.dcerpc import (
    DceRpc5Fault,
    DceRpc5BindNak,
    NDRConfPacketListField,
    NDRConfStrLenField,
    NDRConfStrLenFieldUtf16,
    NDRFullPointerField,
    NDRInt3264EnumField,
    NDRIntField,
    NDRPacket,
    NDRPacketField,
    NDRRecursiveField,
    NDRSerializeType1PacketField,
    NDRShortField,
    NDRSignedIntField,
    NDRSignedShortField,
    NDRUnionField,
    NDRLongField,
)
from scapy.packet import Packet, bind_layers

# [MS-EERR] structures


class EEAString(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRSignedShortField("nLength", None, size_of="pString"),
        NDRFullPointerField(
            NDRConfStrLenField("pString", "", size_is=lambda pkt: pkt.nLength),
            deferred=True,
        ),
    ]


class EEUString(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRSignedShortField("nLength", None, size_of="pString"),
        NDRFullPointerField(
            NDRConfStrLenFieldUtf16("pString", "", size_is=lambda pkt: pkt.nLength),
            deferred=True,
        ),
    ]


class BinaryEEInfo(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRSignedShortField("nSize", None, size_of="pBlob"),
        NDRFullPointerField(
            NDRConfStrLenField("pBlob", "", size_is=lambda pkt: pkt.nSize),
            deferred=True,
        ),
    ]


class ExtendedErrorParamTypesInternal(IntEnum):
    eeptiAnsiString = 1
    eeptiUnicodeString = 2
    eeptiLongVal = 3
    eeptiShortValue = 4
    eeptiPointerValue = 5
    eeptiNone = 6
    eeptiBinary = 7


class ExtendedErrorParam(NDRPacket):
    ALIGNMENT = (8, 8)
    fields_desc = [
        NDRInt3264EnumField("Type", 0, ExtendedErrorParamTypesInternal),
        NDRUnionField(
            [
                (
                    NDRPacketField("value", EEAString(), EEAString),
                    (
                        (lambda pkt: getattr(pkt, "Type", None) == 1),
                        (lambda _, val: val.tag == 1),
                    ),
                ),
                (
                    NDRPacketField("value", EEUString(), EEUString),
                    (
                        (lambda pkt: getattr(pkt, "Type", None) == 2),
                        (lambda _, val: val.tag == 2),
                    ),
                ),
                (
                    NDRIntField("value", 0),
                    (
                        (lambda pkt: getattr(pkt, "Type", None) == 3),
                        (lambda _, val: val.tag == 3),
                    ),
                ),
                (
                    NDRSignedShortField("value", 0),
                    (
                        (lambda pkt: getattr(pkt, "Type", None) == 4),
                        (lambda _, val: val.tag == 4),
                    ),
                ),
                (
                    NDRLongField("value", 0),
                    (
                        (lambda pkt: getattr(pkt, "Type", None) == 5),
                        (lambda _, val: val.tag == 5),
                    ),
                ),
                (
                    StrFixedLenField("value", "", length=0),
                    (
                        (lambda pkt: getattr(pkt, "Type", None) == 6),
                        (lambda _, val: val.tag == 6),
                    ),
                ),
                (
                    NDRPacketField("value", BinaryEEInfo(), BinaryEEInfo),
                    (
                        (lambda pkt: getattr(pkt, "Type", None) == 7),
                        (lambda _, val: val.tag == 7),
                    ),
                ),
            ],
            StrFixedLenField("value", "", length=0),
            align=(2, 8),
            switch_fmt=("H", "I"),
        ),
    ]


class EEComputerNamePresent(IntEnum):
    eecnpPresent = 1
    eecnpNotPresent = 2


class EEComputerName(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRInt3264EnumField("Type", 0, EEComputerNamePresent),
        NDRUnionField(
            [
                (
                    NDRPacketField("value", EEUString(), EEUString),
                    (
                        (lambda pkt: getattr(pkt, "Type", None) == 1),
                        (lambda _, val: val.tag == 1),
                    ),
                ),
                (
                    StrFixedLenField("value", "", length=0),
                    (
                        (lambda pkt: getattr(pkt, "Type", None) == 2),
                        (lambda _, val: val.tag == 2),
                    ),
                ),
            ],
            StrFixedLenField("value", "", length=0),
            align=(2, 8),
            switch_fmt=("H", "I"),
        ),
    ]


class ExtendedErrorInfo(NDRPacket):
    ALIGNMENT = (8, 8)
    DEPORTED_CONFORMANTS = ["Params"]
    fields_desc = [
        NDRRecursiveField("Next"),
        NDRPacketField("ComputerName", EEComputerName(), EEComputerName),
        NDRIntField("ProcessID", 0),
        NDRLongField("TimeStamp", 0),
        NDRIntField("GeneratingComponent", 0),
        NDRIntField("Status", 0),
        NDRShortField("DetectionLocation", 0),
        NDRShortField("Flags", 0),
        NDRSignedShortField("nLen", None, size_of="Params"),
        NDRConfPacketListField(
            "Params",
            [],
            ExtendedErrorParam,
            size_is=lambda pkt: pkt.nLen,
            conformant_in_struct=True,
        ),
    ]


# Encapsulation packets

# https://learn.microsoft.com/en-us/windows/win32/rpc/understanding-extended-error-information

EERR_GENERATING_COMPONENT = {
    1: "Application",  # The component owning the manager routine for the particular RPC call
    2: "Runtime",  # The RPC run time
    3: "Security Provider",  # The security provider for this call.
    4: "NPFS",  # The NPFS file system
    5: "RDR",  # The Redirector
    6: "NMP",  # The named pipe system. This can be either NPFS or RDR, but in many cases the RPC run time does not know who performed the requested the operation, and in such cases NMP is returned.
    7: "IO",  # The IO system or a driver used by the IO system. This can be either NPFS, RDR, or a Winsock provider.
    8: "Winsock",  # The Winsock provider
    9: "Authz code",  # The Authorization APIs.
    10: "LPC",  # The Local Procedure Call facility.
}

EERR_FLAGS = {
    1: "EEInfoPreviousRecordsMissing",
    2: "EEInfoNextRecordsMissing",
}

# https://learn.microsoft.com/en-us/windows/win32/rpc/extended-error-information-detection-locations

EERR_DETECTION_LOCATIONS = {
    10: "DealWithLRPCRequest10",
    11: "DealWithLRPCRequest20",
    12: "WithLRPCRequest30",
    13: "WithLRPCRequest40",
    20: "LrpcMessageToRpcMessage10",
    21: "LrpcMessageToRpcMessage20",
    22: "LrpcMessageToRpcMessage30",
    30: "DealWithRequestMessage10",
    31: "DealWithRequestMessage20",
    32: "DealWithRequestMessage30",
    40: "CheckSecurity10",
    50: "DealWithBindMessage10",
    51: "DealWithBindMessage20",
    52: "DealWithBindMessage30",
    53: "DealWithBindMessage40",
    54: "DealWithBindMessage50",
    55: "DealWithBindMessage60",
    60: "FindServerCredentials10",
    61: "FindServerCredentials20",
    62: "FindServerCredentials30",
    70: "AcceptFirstTime10",
    71: "AcceptThirdLeg10",
    72: "AcceptThirdLeg20",
    73: "AcceptFirstTime20",
    74: "AcceptThirdLeg40",
    80: "AssociationRequested10",
    81: "AssociationRequested20",
    82: "AssociationRequested30",
    90: "CompleteSecurityToken10",
    91: "CompleteSecurityToken20",
    100: "AcquireCredentialsForClient10",
    101: "AcquireCredentialsForClient20",
    102: "AcquireCredentialsForClient30",
    110: "InquireDefaultPrincName10",
    111: "InquireDefaultPrincName20",
    120: "SignOrSeal10",
    130: "VerifyOrUnseal10",
    131: "VerifyOrUnseal20",
    140: "InitializeFirstTime10",
    141: "InitializeFirstTime20",
    142: "InitializeFirstTime30",
    150: "InitializeThirdLeg10",
    151: "InitializeThirdLeg20",
    152: "InitializeThirdLeg30",
    153: "InitializeThirdLeg40",
    154: "InitializeThirdLeg50",
    155: "InitializeThirdLeg60",
    160: "ImpersonateClient10",
    170: "DispatchToStub10",
    171: "DispatchToStub20",
    180: "DispatchToStubWorker10",
    181: "DispatchToStubWorker20",
    182: "DispatchToStubWorker30",
    183: "DispatchToStubWorker40",
    190: "NMPOpen10",
    191: "NMPOpen20",
    192: "NMPOpen30",
    193: "NMPOpen40",
    200: "NMPSyncSend10",
    210: "NMPSyncSendReceive10",
    220: "NMPSyncSendReceive20",
    221: "NMPSyncSendReceive30",
    230: "COSend10",
    240: "COSubmitRead10",
    250: "COSubmitSyncRead10",
    251: "COSubmitSyncRead20",
    260: "COSyncRecv10",
    270: "WSCheckForShutdowns10",
    271: "WSCheckForShutdowns20",
    272: "WSCheckForShutdowns30",
    273: "WSCheckForShutdowns40",
    274: "WSCheckForShutdowns50",
    280: "WSSyncSend10",
    281: "WSSyncSend20",
    282: "WSSyncSend30",
    290: "WSSyncRecv10",
    291: "WSSyncRecv20",
    292: "WSSyncRecv30",
    300: "WSServerListenCommon10",
    301: "WSServerListenCommon20",
    302: "WSServerListenCommon30",
    310: "WSOpen10",
    311: "WSOpen20",
    312: "WSOpen30",
    313: "WSOpen40",
    314: "WSOpen50",
    315: "WSOpen60",
    316: "WSOpen70",
    317: "WSOpen80",
    318: "WSOpen90",
    320: "NextAddress10",
    321: "NextAddress20",
    322: "NextAddress30",
    323: "NextAddress40",
    330: "WSBind10",
    331: "WSBind20",
    332: "WSBind30",
    333: "WSBind40",
    334: "WSBind50",
    335: "WSBind45",
    340: "IPBuildAddressVector10",
    350: "GetStatusForTimeout10",
    351: "GetStatusForTimeout20",
    360: "OSF_CCONNECTION__SendFragment10",
    361: "OSF_CCONNECTION__SendFragment20",
    370: "OSF_CCALL__ReceiveReply10",
    371: "OSF_CCALL__ReceiveReply20",
    380: "OSF_CCALL__FastSendReceive10",
    381: "OSF_CCALL__FastSendReceive20",
    382: "OSF_CCALL__FastSendReceive30",
    390: "LRPC_BINDING_HANDLE__AllocateCCall10",
    391: "LRPC_BINDING_HANDLE__AllocateCCall20",
    400: "LRPC_ADDRESS__ServerSetupAddress10",
    410: "LRPC_ADDRESS__HandleInvalidAssociationReference10",
    420: "InitializeAuthzSupportIfNecessary10",
    421: "InitializeAuthzSupportIfNecessary20",
    430: "CreateDummyResourceManagerIfNecessary10",
    431: "CreateDummyResourceManagerIfNecessary20",
    440: "LRPC_SCALL__GetAuthorizationContext10",
    441: "LRPC_SCALL__GetAuthorizationContext20",
    442: "LRPC_SCALL__GetAuthorizationContext30",
    450: "SCALL__DuplicateAuthzContext10",
    460: "SCALL__CreateAndSaveAuthzContextFromToken10",
    470: "SECURITY_CONTEXT__GetAccessToken10",
    471: "SECURITY_CONTEXT__GetAccessToken20",
    480: "OSF_SCALL__GetAuthorizationContext10",
    500: "EpResolveEndpoint10",
    501: "EpResolveEndpoint20",
    510: "OSF_SCALL__GetBuffer10",
    520: "LRPC_SCALL__ImpersonateClient10",
    530: "SetMaximumLengths10",
    540: "LRPC_CASSOCIATION__ActuallyDoBinding10",
    541: "LRPC_CASSOCIATION__ActuallyDoBinding20",
    542: "LRPC_CASSOCIATION__ActuallyDoBinding30",
    543: "LRPC_CASSOCIATION__ActuallyDoBinding40",
    550: "LRPC_CASSOCIATION__CreateBackConnection10",
    551: "LRPC_CASSOCIATION__CreateBackConnection20",
    552: "LRPC_CASSOCIATION__CreateBackConnection30",
    560: "LRPC_CASSOCIATION__OpenLpcPort10",
    561: "LRPC_CASSOCIATION__OpenLpcPort20",
    562: "LRPC_CASSOCIATION__OpenLpcPort30",
    563: "LRPC_CASSOCIATION__OpenLpcPort40",
    570: "RegisterEntries10",
    571: "RegisterEntries20",
    580: "NDRSContextUnmarshall2_10",
    581: "NDRSContextUnmarshall2_20",
    582: "NDRSContextUnmarshall2_30",
    583: "NDRSContextUnmarshall2_40",
    584: "NDRSContextUnmarshall2_50",
    590: "NDRSContextMarshall2_10",
    600: "WinsockDatagramSend10",
    601: "WinsockDatagramSend20",
    610: "WinsockDatagramReceive10",
    620: "WinsockDatagramSubmitReceive10",
    630: "DG_CCALL__CancelAsyncCall10",
    640: "DG_CCALL__DealWithTimeout10",
    641: "DG_CCALL__DealWithTimeout20",
    642: "DG_CCALL__DealWithTimeout30",
    650: "DG_CCALL__DispatchPacket10",
    660: "DG_CCALL__ReceiveSinglePacket10",
    661: "DG_CCALL__ReceiveSinglePacket20",
    662: "DG_CCALL__ReceiveSinglePacket30",
    670: "WinsockDatagramResolve10",
    680: "WinsockDatagramCreate10",
    690: "TCP_QueryLocalAddress10",
    691: "TCP_QueryLocalAddress20",
    700: "OSF_CASSOCIATION__ProcessBindAckOrNak10",
    701: "OSF_CASSOCIATION__ProcessBindAckOrNak20",
    710: "MatchMsPrincipalName10",
    720: "CompareRdnElement10",
    730: "MatchFullPathPrincipalName10",
    731: "MatchFullPathPrincipalName20",
    732: "MatchFullPathPrincipalName30",
    733: "MatchFullPathPrincipalName40",
    734: "MatchFullPathPrincipalName50",
    740: "RpcCertGeneratePrincipalName10",
    741: "RpcCertGeneratePrincipalName20",
    742: "RpcCertGeneratePrincipalName30",
    750: "RpcCertVerifyContext10",
    751: "RpcCertVerifyContext20",
    752: "RpcCertVerifyContext30",
    753: "RpcCertVerifyContext40",
    761: "OSF_BINDING_HANDLE__NegotiateTransferSyntax10",
    # END OF DOC
    # below is reverse engineered
    770: "RpcpErrorAddRecord",
    780: "RpcpLookupAccountSid",
    800: "OSF_SCONNECTION__AcceptFirstTime",
    810: "OSF_SCONNECTION__AcceptThirdLeg",
    820: "OSF_BINDING_HANDLE__AcquireTokenForTransport",
    821: "OSF_BINDING_HANDLE__AcquireCredentialsForTokenIfNecessary",
    835: "LRPC_CASSOCIATION__CompleteBind",
    840: "LRPC_BASE_BINDING_HANDLE__BaseBindingCopy",
    860: "LRPC_BASE_BINDING_HANDLE__SetAuthInformation",
    861: "LRPC_BASE_BINDING_HANDLE__SetAuthInformation",
    862: "LRPC_BASE_BINDING_HANDLE__SetAuthInformation",
    864: "LRPC_BASE_BINDING_HANDLE__SetAuthInformation",
    865: "LRPC_BASE_BINDING_HANDLE__SetAuthInformation",
    867: "LRPC_BASE_BINDING_HANDLE__SetAuthInformation",
    868: "LRPC_BASE_BINDING_HANDLE__SetAuthInformation",
    869: "LRPC_BASE_BINDING_HANDLE__SetAuthInformation",
    880: "LRPC_BASE_BINDING_HANDLE__ResolveEndpoint",
    881: "LRPC_BASE_BINDING_HANDLE__ResolveEndpoint",
    882: "LRPC_BASE_BINDING_HANDLE__ResolveEndpoint",
    883: "LRPC_BASE_BINDING_HANDLE__ResolveEndpoint",
    900: "LRPC_BASE_BINDING_HANDLE__SubmitResolveEndpointRequest",
    910: "LRPC_BASE_BINDING_HANDLE__NormalizeServerSid",
    912: "LRPC_BASE_BINDING_HANDLE__NormalizeServerSid",
    913: "LRPC_BASE_BINDING_HANDLE__NormalizeServerSid",
    920: "LRPC_BASE_BINDING_HANDLE__DriveStateForward",
    921: "LRPC_BASE_BINDING_HANDLE__DriveStateForward",
    930: "LRPC_BINDING_HANDLE__PrepareBindingHandle",
    931: "LRPC_BINDING_HANDLE__PrepareBindingHandle",
    946: "?0LRPC_FAST_BINDING_HANDLE",
    950: "LRPC_BASE_CCALL__AsyncReceive",
    960: "LRPC_BINDING_HANDLE__NegotiateTransferSyntax",
    972: "LRPC_BASE_CCALL__UnpackResponse",
    974: "LRPC_BASE_CCALL__UnpackResponse",
    978: "LRPC_BASE_CCALL__UnpackResponse",
    980: "LRPC_BASE_CCALL__HandleReply",
    990: "LRPC_BASE_CCALL__HandleReply",
    1000: "LRPC_BASE_CCALL__AlpcSend",
    1010: "LRPC_BASE_CCALL__DoSendReceive",
    1020: "LRPC_BASE_CCALL__GetBuffer",
    1030: "LRPC_BASE_CCALL__DoAsyncSend",
    1040: "LRPC_BIND_CCALL__NotifyBHStateChange",
    1050: "LRPC_BIND_CCALL__AttemptRetry",
    1060: "CLIENT_IO_PROVIDER__SyncWait",
    1070: "CLIENT_IO_PROVIDER__Register",
    1100: "AlpcAllocateSectionAndView",
    1120: "CaptureThreadToken",
    1140: "RpcpDuplicateTokenEx",
    1150: "GetMachineAccountSidWorker",
    1160: "LRPC_BASE_BINDING_HANDLE__DriveStateForward",
    1170: "LPC_NORMALIZED_SID__IterateAndVerify",
    1100: "AlpcAllocateSectionAndView",
    1120: "CaptureThreadToken",
    1140: "RpcpDuplicateTokenEx",
    1150: "GetMachineAccountSidWorker",
    1160: "LRPC_BASE_BINDING_HANDLE__DriveStateForward",
    1170: "LPC_NORMALIZED_SID__IterateAndVerify",
    1171: "LPC_NORMALIZED_SID__IterateAndVerify",
    1180: "LRPC_SASSOCIATION__GetClientName",
    1190: "LRPC_SASSOCIATION__AddBinding",
    1200: "AlpcAllocateView",
    1210: "LRPC_SASSOCIATION__ImpersonateClient",
    1215: "LRPC_SASSOCIATION__ImpersonateClientContainer",
    1230: "LRPC_SCALL__SaveClientToken",
    1232: "LRPC_SCALL__SaveClientToken",
    1240: "LRPC_SCONTEXT__GetUserNameW",
    1260: "LRPC_SCONTEXT__LookupUser",
    1270: "LRPC_BINDING_HANDLE__NegotiateTransferSyntax",
    1280: "LRPC_SCALL__SendReceive",
    1290: "LRPC_CCALL__HandleCallbackSequence",
    1300: "LRPC_SCALL__QueueOrDispatchCall",
    1310: "LRPC_ADDRESS__GetCurrentModifiedId",
    1320: "RPC_INTERFACE__DoSecurityCallbackHelper",
    1321: "RPC_INTERFACE__DoSecurityCallbackHelper",
    1322: "RPC_INTERFACE__DoSecurityCallbackHelper",
    1324: "RPC_INTERFACE__EnforceInterfaceSecurityDescriptor",
    1325: "RPC_INTERFACE__EnforceInterfaceSecurityDescriptor",
    1326: "RPC_INTERFACE__EnforceInterfaceSecurityDescriptor",
    1330: "SCALL__CompleteAsyncSecurityCallback",
    1340: "LRPC_BASE_CCALL__ReallocPipeBuffer",
    1360: "LRPC_ADDRESS__GetClientSid",
    1440: "RpcpErrorAddRecord",
    1441: "RpcpErrorAddRecord",
    1442: "TCPOrHTTP_Open",
    1450: "I_RpcRecordCalloutFailure",
    1451: "RpcpErrorAddRecord",
    1452: "I_RpcRecordCalloutFailure",
    1460: "OSF_CCONNECTION__SendBindPacket",
    1461: "OSF_CCONNECTION__SendBindPacket",
    1462: "OSF_CCONNECTION__SendBindPacket",
    1463: "OSF_CCONNECTION__SendBindPacket",
    1464: "OSF_CCONNECTION__SendBindPacket",
    1465: "OSF_CCONNECTION__SendBindPacket",
    1466: "OSF_CCONNECTION__SendBindPacket",
    1467: "OSF_CCONNECTION__SendBindPacket",
    1468: "OSF_CCONNECTION__SendBindPacket",
    1469: "OSF_CCONNECTION__SendBindPacket",
    1471: "LRPC_BASE_CCALL__HandlePipeChunk",
    1474: "LRPC_BASE_CCALL__HandlePipeChunk",
    1480: "LRPC_CASSOCIATION__AlpcCreateReservedMessage",
    1491: "LRPC_CASSOCIATION__AskForReservedMessage",
    1492: "LRPC_CASSOCIATION__AskForReservedMessage",
    1500: "LRPC_BASE_CCALL__CancelAsyncCall",
    1502: "LRPC_BASE_CCALL__HandlePipeFault",
    1510: "LRPC_BASE_CCALL__HandlePipePull",
    1520: "LRPC_SCALL__NotifyAssociationClosePending",
    1530: "LRPC_SCALL__AbortAsyncCall",
    1540: "BindToEpMapper",
    1550: "LRPC_BASE_CCALL__Send",
    1560: "LRPC_BASE_CCALL__Receive",
    1570: "LRPC_SASSOCIATION__NotifyAllActiveCalls",
    1580: "LRPC_SASSOCIATION__CleanupSparsePipes",
    1590: "LRPC_CCALL__CallbackSendReceive",
    1600: "LRPC_CALLBACK__SendReceiveLoop",
    1601: "LRPC_CALLBACK__SendReceiveLoop",
    1602: "LRPC_CALLBACK__SendReceiveLoop",
    1610: "OSF_SCALL__DoSecurityCallbackAndAccessCheck",
    1611: "OSF_SCALL__DoSecurityCallbackAndAccessCheck",
    1612: "OSF_SCALL__DoSecurityCallbackAndAccessCheck",
    1613: "OSF_SCALL__DoSecurityCallbackAndAccessCheck",
    1614: "OSF_SCALL__DoSecurityCallbackAndAccessCheck",
    1616: "OSF_SCALL__DoSecurityCallbackAndAccessCheck",
    1617: "OSF_SCALL__DoSecurityCallbackAndAccessCheck",
    1619: "OSF_SCALL__DoSecurityCallbackAndAccessCheck",
    1641: "LRPC_SCAUSAL_FLOW__MaybeQueueCall",
    1642: "LRPC_SCAUSAL_FLOW__MaybeQueueCall",
    1650: "LRPC_FAST_BIND_CCALL__ActualCancelCall",
    1660: "LRPC_FAST_BINDING_HANDLE__Bind",
    1670: "LRPC_CAUSAL_FLOW__SendNextCalls",
    1700: "CO_ConnectionThreadPoolCallback",
    1701: "CO_ConnectionThreadPoolCallback",
    1704: "CO_AddressThreadPoolCallback",
    1705: "CO_AddressThreadPoolCallback",
    1710: "OSF_CCONNECTION__ConnectionAborted",
    1720: "OSF_CCONNECTION__ProcessReceiveComplete",
    1730: "OSF_CCONNECTION__ProcessSendComplete",
    1740: "OSF_BINDING_HANDLE__AllocateCCall",
    1741: "OSF_BINDING_HANDLE__AllocateCCall",
    1750: "ProcessFaultPacket",
    3050: "OSF_CCONNECTION__AddCall",
    3080: "SECURITY_CONTEXT__GetWireIdForSnego",
    3081: "SECURITY_CONTEXT__GetWireIdForSnego",
    4000: "OSF_SCALL__FwFilter",
    4001: "OSF_SCALL__FwFilter",
    4002: "OSF_SCALL__FwFilter",
    4020: "LRPC_SCALL__ReallocPipeBuffer",
    4030: "LRPC_BASE_CCALL__HandleCancelMessage",
    4040: "LRPC_SCAUSAL_FLOW__RecordGapInFlow",
    4060: "NMP_BuildDefaultDaclForPipe",
    4061: "NMP_BuildDefaultDaclForPipe",
    4070: "NMP_SetSecurity",
    4072: "NMP_SetSecurity",
    4073: "NMP_BuildDefaultDaclForPipe",
    4077: "NMP_SetSecurity",
    4078: "NMP_SetSecurity",
    4200: "WS_ClientBind",
    4210: "OSF_SCALL__ProcessReceivedPDU",
    4222: "OSF_SCALL__EatAuthInfoFromPacket",
    4230: "RpcpConvertToLongRunning",
    4240: "LRPC_BASE_BINDING_HANDLE__EnableAsyncIfNecessary",
    4253: "LRPC_SYSTEM_HANDLE_DATA__AddSystemHandle",
    4256: "LRPC_SYSTEM_HANDLE_DATA__GetSystemHandle",
    4257: "LRPC_SYSTEM_HANDLE_DATA__GetSystemHandle",
    4272: "RpcpSystemHandleTypeSpecificWork",
    4273: "RpcpSystemHandleTypeSpecificWork",
    4274: "RpcpSystemHandleTypeSpecificWork",
    4291: "HVSOCKET_SetSocketOptions",
}


class DceRpc5ExtendedErrorInfo(Packet):
    fields_desc = [
        NDRSerializeType1PacketField(
            "extended_error",
            ExtendedErrorInfo(),
            ExtendedErrorInfo,
        )
    ]

    def show(self) -> None:
        """
        Print stacktrace
        """
        # Get a list of ErrorInfo
        cur = self.extended_error.value
        errors = [cur]
        while cur and cur.Next:
            cur = cur.Next.value
            errors.append(cur)
        # Concatenate the ErrorInfos
        timefld = UTCTimeField(
            "",
            None,
            fmt="<Q",
            epoch=[1601, 1, 1, 0, 0, 0],
            custom_scaling=1e7,
            strf="%d/%m/%Y %H:%M:%S.%f",
        )
        print("# Extended Error Information")
        for err in errors:
            print("PID:", err.ProcessID, "-", timefld.i2repr(None, err.TimeStamp))
            if err.ComputerName.Type == EEComputerNamePresent.eecnpPresent:
                print(
                    " | ComputerName:",
                    err.ComputerName.value.value.valueof("pString").decode("utf-16"),
                )
            print(
                " | Generating Component:",
                EERR_GENERATING_COMPONENT.get(
                    err.GeneratingComponent, err.GeneratingComponent
                ),
            )
            print(" | Status:", err.Status)
            print(
                " | DetectionLocation:",
                EERR_DETECTION_LOCATIONS.get(
                    err.DetectionLocation, err.DetectionLocation
                ),
            )
            print(" | Flags", EERR_FLAGS.get(err.Flags, err.Flags))
            print(
                " | Params: ",
                repr([(x.sprintf("%Type%"), x.value.value) for x in err.Params]),
            )


# Bind to fault PDU
bind_layers(DceRpc5Fault, DceRpc5ExtendedErrorInfo, reserved=1)
# Bind to nak PDU
bind_layers(DceRpc5BindNak, DceRpc5ExtendedErrorInfo, signature=uuid.UUID("90740320-fad0-11d3-82d7-009027b130ab"))



### ms-oxabref.idl
"""
RPC definitions for the following interfaces:
- rfri (v1.0): 1544f5e0-613c-11d1-93df-00c04fd7bd09
This file is auto-generated by midl-to-scapy, do not modify.
"""

import uuid


from scapy.layers.dcerpc import (
    NDRPacket,
    DceRpcOp,
    NDRConfVarStrLenField,
    NDRConfVarStrNullField,
    NDRFullPointerField,
    NDRIntField,
    register_dcerpc_interface,
)


class RfrGetNewDSA_Request(NDRPacket):
    fields_desc = [
        NDRIntField("ulFlags", 0),
        NDRConfVarStrNullField("pUserDN", ""),
        NDRFullPointerField(
            NDRFullPointerField(NDRConfVarStrNullField("ppszUnused", ""))
        ),
        NDRFullPointerField(
            NDRFullPointerField(NDRConfVarStrNullField("ppszServer", ""))
        ),
    ]


class RfrGetNewDSA_Response(NDRPacket):
    fields_desc = [
        NDRFullPointerField(
            NDRFullPointerField(NDRConfVarStrNullField("ppszUnused", ""))
        ),
        NDRFullPointerField(
            NDRFullPointerField(NDRConfVarStrNullField("ppszServer", ""))
        ),
        NDRIntField("status", 0),
    ]


class RfrGetFQDNFromServerDN_Request(NDRPacket):
    fields_desc = [
        NDRIntField("ulFlags", 0),
        NDRIntField("cbMailboxServerDN", None, size_of="szMailboxServerDN"),
        NDRConfVarStrLenField(
            "szMailboxServerDN", "", size_is=lambda pkt: pkt.cbMailboxServerDN
        ),
    ]


class RfrGetFQDNFromServerDN_Response(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullField("ppszServerFQDN", "")),
        NDRIntField("status", 0),
    ]


RFRI_OPNUMS = {
    0: DceRpcOp(RfrGetNewDSA_Request, RfrGetNewDSA_Response),
    1: DceRpcOp(RfrGetFQDNFromServerDN_Request, RfrGetFQDNFromServerDN_Response),
}
register_dcerpc_interface(
    name="rfri",
    uuid=uuid.UUID("1544f5e0-613c-11d1-93df-00c04fd7bd09"),
    version="1.0",
    opnums=RFRI_OPNUMS,
)



### ms-fax_faxdatatypes.idl
"""
RPC definitions for the following interfaces:
-
This file is auto-generated by midl-to-scapy, do not modify.
"""

from enum import IntEnum
import uuid


from scapy.layers.dcerpc import (
    NDRPacket,
    NDRConfFieldListField,
    NDRConfVarStrNullField,
    NDRConfVarStrNullFieldUtf16,
    NDRFieldListField,
    NDRFullPointerField,
    NDRInt3264EnumField,
    NDRInt3264Field,
    NDRIntField,
    NDRPacketField,
    NDRShortField,
    NDRSignedIntField,
    NDRUnionField,
)


class FAX_COVERPAGE_INFO_EXW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRIntField("dwCoverPageFormat", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrCoverPageFileName", ""), deferred=True
        ),
        NDRSignedIntField("bServerBased", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrNote", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrSubject", ""), deferred=True
        ),
    ]


class PFAX_COVERPAGE_INFO_EXW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRIntField("dwCoverPageFormat", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrCoverPageFileName", ""), deferred=True
        ),
        NDRSignedIntField("bServerBased", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrNote", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrSubject", ""), deferred=True
        ),
    ]


class LPCFAX_COVERPAGE_INFO_EXW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRIntField("dwCoverPageFormat", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrCoverPageFileName", ""), deferred=True
        ),
        NDRSignedIntField("bServerBased", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrNote", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrSubject", ""), deferred=True
        ),
    ]


class SYSTEMTIME(NDRPacket):
    ALIGNMENT = (2, 2)
    fields_desc = [
        NDRShortField("wYear", 0),
        NDRShortField("wMonth", 0),
        NDRShortField("wDayOfWeek", 0),
        NDRShortField("wDay", 0),
        NDRShortField("wHour", 0),
        NDRShortField("wMinute", 0),
        NDRShortField("wSecond", 0),
        NDRShortField("wMilliseconds", 0),
    ]


class FAX_JOB_PARAMW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("SizeOfStruct", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("RecipientNumber", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("RecipientName", ""), deferred=True
        ),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("Tsid", ""), deferred=True),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("SenderName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("SenderCompany", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("SenderDept", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("BillingCode", ""), deferred=True
        ),
        NDRIntField("ScheduleAction", 0),
        NDRPacketField("ScheduleTime", SYSTEMTIME(), SYSTEMTIME),
        NDRIntField("DeliveryReportType", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("DeliveryReportAddress", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("DocumentName", ""), deferred=True
        ),
        NDRIntField("CallHandle", 0),
        NDRFieldListField(
            "Reserved", [], NDRInt3264Field("", 0), length_is=lambda _: 3
        ),
    ]


class PFAX_JOB_PARAMW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("SizeOfStruct", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("RecipientNumber", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("RecipientName", ""), deferred=True
        ),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("Tsid", ""), deferred=True),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("SenderName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("SenderCompany", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("SenderDept", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("BillingCode", ""), deferred=True
        ),
        NDRIntField("ScheduleAction", 0),
        NDRPacketField("ScheduleTime", SYSTEMTIME(), SYSTEMTIME),
        NDRIntField("DeliveryReportType", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("DeliveryReportAddress", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("DocumentName", ""), deferred=True
        ),
        NDRIntField("CallHandle", 0),
        NDRFieldListField(
            "Reserved", [], NDRInt3264Field("", 0), length_is=lambda _: 3
        ),
    ]


class FAX_ENUM_DEVICE_RECEIVE_MODE(IntEnum):
    FAX_DEVICE_RECEIVE_MODE_OFF = 0
    FAX_DEVICE_RECEIVE_MODE_AUTO = 1
    FAX_DEVICE_RECEIVE_MODE_MANUAL = 2


class FAX_ENUM_GROUP_STATUS(IntEnum):
    FAX_GROUP_STATUS_ALL_DEV_VALID = 0
    FAX_GROUP_STATUS_EMPTY = 1
    FAX_GROUP_STATUS_ALL_DEV_NOT_VALID = 2
    FAX_GROUP_STATUS_SOME_DEV_NOT_VALID = 3


class FAX_ENUM_MESSAGE_FOLDER(IntEnum):
    FAX_MESSAGE_FOLDER_INBOX = 0
    FAX_MESSAGE_FOLDER_SENTITEMS = 1
    FAX_MESSAGE_FOLDER_QUEUE = 2


class FAX_ENUM_PERSONAL_PROF_TYPES(IntEnum):
    RECIPIENT_PERSONAL_PROF = 1
    SENDER_PERSONAL_PROF = 2


class FAX_ENUM_PRIORITY_TYPE(IntEnum):
    FAX_PRIORITY_TYPE_LOW = 0
    FAX_PRIORITY_TYPE_NORMAL = 1
    FAX_PRIORITY_TYPE_HIGH = 2


class FAX_ENUM_SMTP_AUTH_OPTIONS(IntEnum):
    FAX_SMTP_AUTH_ANONYMOUS = 0
    FAX_SMTP_AUTH_BASIC = 1
    FAX_SMTP_AUTH_NTLM = 2


class PRODUCT_SKU_TYPE(IntEnum):
    PRODUCT_SKU_UNKNOWN = 0
    PRODUCT_SKU_PERSONAL = 1
    PRODUCT_SKU_PROFESSIONAL = 2
    PRODUCT_SKU_SERVER = 4
    PRODUCT_SKU_ADVANCED_SERVER = 8
    PRODUCT_SKU_DATA_CENTER = 16
    PRODUCT_SKU_DESKTOP_EMBEDDED = 32
    PRODUCT_SKU_SERVER_EMBEDDED = 64
    PRODUCT_SKU_WEB_SERVER = 128


class FAX_ENUM_CONFIG_OPTION(IntEnum):
    FAX_CONFIG_OPTION_ALLOW_PERSONAL_CP = 0
    FAX_CONFIG_OPTION_QUEUE_STATE = 1
    FAX_CONFIG_OPTION_ALLOWED_RECEIPTS = 2
    FAX_CONFIG_OPTION_INCOMING_FAXES_PUBLIC = 3


class FAX_TIME(NDRPacket):
    ALIGNMENT = (2, 2)
    fields_desc = [NDRShortField("Hour", 0), NDRShortField("Minute", 0)]


class PFAX_TIME(NDRPacket):
    ALIGNMENT = (2, 2)
    fields_desc = [NDRShortField("Hour", 0), NDRShortField("Minute", 0)]


class FAX_RECEIPTS_CONFIGW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRIntField("dwAllowedReceipts", 0),
        NDRInt3264EnumField("SMTPAuthOption", 0, FAX_ENUM_SMTP_AUTH_OPTIONS),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrReserved", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrSMTPServer", ""), deferred=True
        ),
        NDRIntField("dwSMTPPort", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrSMTPFrom", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrSMTPUserName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrSMTPPassword", ""), deferred=True
        ),
        NDRSignedIntField("bIsToUseForMSRouteThroughEmailMethod", 0),
    ]


class PFAX_RECEIPTS_CONFIGW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRIntField("dwAllowedReceipts", 0),
        NDRInt3264EnumField("SMTPAuthOption", 0, FAX_ENUM_SMTP_AUTH_OPTIONS),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrReserved", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrSMTPServer", ""), deferred=True
        ),
        NDRIntField("dwSMTPPort", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrSMTPFrom", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrSMTPUserName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrSMTPPassword", ""), deferred=True
        ),
        NDRSignedIntField("bIsToUseForMSRouteThroughEmailMethod", 0),
    ]


class FAX_CONFIGURATIONW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("SizeOfStruct", 0),
        NDRIntField("Retries", 0),
        NDRIntField("RetryDelay", 0),
        NDRIntField("DirtyDays", 0),
        NDRSignedIntField("Branding", 0),
        NDRSignedIntField("UseDeviceTsid", 0),
        NDRSignedIntField("ServerCp", 0),
        NDRSignedIntField("PauseServerQueue", 0),
        NDRPacketField("StartCheapTime", FAX_TIME(), FAX_TIME),
        NDRPacketField("StopCheapTime", FAX_TIME(), FAX_TIME),
        NDRSignedIntField("ArchiveOutgoingFaxes", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("ArchiveDirectory", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("ProfileName", ""), deferred=True
        ),
    ]


class PFAX_CONFIGURATIONW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("SizeOfStruct", 0),
        NDRIntField("Retries", 0),
        NDRIntField("RetryDelay", 0),
        NDRIntField("DirtyDays", 0),
        NDRSignedIntField("Branding", 0),
        NDRSignedIntField("UseDeviceTsid", 0),
        NDRSignedIntField("ServerCp", 0),
        NDRSignedIntField("PauseServerQueue", 0),
        NDRPacketField("StartCheapTime", FAX_TIME(), FAX_TIME),
        NDRPacketField("StopCheapTime", FAX_TIME(), FAX_TIME),
        NDRSignedIntField("ArchiveOutgoingFaxes", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("ArchiveDirectory", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("ProfileName", ""), deferred=True
        ),
    ]


class FAX_GLOBAL_ROUTING_INFOW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("SizeOfStruct", 0),
        NDRIntField("Priority", 0),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("Guid", ""), deferred=True),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("FriendlyName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("FunctionName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("ExtensionImageName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("ExtensionFriendlyName", ""), deferred=True
        ),
    ]


class PFAX_GLOBAL_ROUTING_INFOW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("SizeOfStruct", 0),
        NDRIntField("Priority", 0),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("Guid", ""), deferred=True),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("FriendlyName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("FunctionName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("ExtensionImageName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("ExtensionFriendlyName", ""), deferred=True
        ),
    ]


class FAX_JOB_PARAM_EXW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRIntField("dwScheduleAction", 0),
        NDRPacketField("tmSchedule", SYSTEMTIME(), SYSTEMTIME),
        NDRIntField("dwReceiptDeliveryType", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrReceiptDeliveryAddress", ""),
            deferred=True,
        ),
        NDRInt3264EnumField("Priority", 0, FAX_ENUM_PRIORITY_TYPE),
        NDRIntField("hCall", 0),
        NDRFieldListField(
            "dwReserved", [], NDRInt3264Field("", 0), length_is=lambda _: 4
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrDocumentName", ""), deferred=True
        ),
        NDRIntField("dwPageCount", 0),
    ]


class PFAX_JOB_PARAM_EXW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRIntField("dwScheduleAction", 0),
        NDRPacketField("tmSchedule", SYSTEMTIME(), SYSTEMTIME),
        NDRIntField("dwReceiptDeliveryType", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrReceiptDeliveryAddress", ""),
            deferred=True,
        ),
        NDRInt3264EnumField("Priority", 0, FAX_ENUM_PRIORITY_TYPE),
        NDRIntField("hCall", 0),
        NDRFieldListField(
            "dwReserved", [], NDRInt3264Field("", 0), length_is=lambda _: 4
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrDocumentName", ""), deferred=True
        ),
        NDRIntField("dwPageCount", 0),
    ]


class LPCFAX_JOB_PARAM_EXW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRIntField("dwScheduleAction", 0),
        NDRPacketField("tmSchedule", SYSTEMTIME(), SYSTEMTIME),
        NDRIntField("dwReceiptDeliveryType", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrReceiptDeliveryAddress", ""),
            deferred=True,
        ),
        NDRInt3264EnumField("Priority", 0, FAX_ENUM_PRIORITY_TYPE),
        NDRIntField("hCall", 0),
        NDRFieldListField(
            "dwReserved", [], NDRInt3264Field("", 0), length_is=lambda _: 4
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrDocumentName", ""), deferred=True
        ),
        NDRIntField("dwPageCount", 0),
    ]


class RPC_FAX_OUTBOUND_ROUTING_GROUPW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrGroupName", ""), deferred=True
        ),
        NDRIntField("dwNumDevices", None, size_of="lpdwDevices"),
        NDRFullPointerField(
            NDRConfFieldListField(
                "lpdwDevices", [], NDRIntField, size_is=lambda pkt: pkt.dwNumDevices
            ),
            deferred=True,
        ),
        NDRInt3264EnumField("Status", 0, FAX_ENUM_GROUP_STATUS),
    ]


class PRPC_FAX_OUTBOUND_ROUTING_GROUPW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrGroupName", ""), deferred=True
        ),
        NDRIntField("dwNumDevices", None, size_of="lpdwDevices"),
        NDRFullPointerField(
            NDRConfFieldListField(
                "lpdwDevices", [], NDRIntField, size_is=lambda pkt: pkt.dwNumDevices
            ),
            deferred=True,
        ),
        NDRInt3264EnumField("Status", 0, FAX_ENUM_GROUP_STATUS),
    ]


class FAX_PORT_INFO(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("SizeOfStruct", 0),
        NDRIntField("DeviceId", 0),
        NDRIntField("State", 0),
        NDRIntField("Flags", 0),
        NDRIntField("Rings", 0),
        NDRIntField("Priority", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("DeviceName", ""), deferred=True
        ),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("Tsid", ""), deferred=True),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("Csid", ""), deferred=True),
    ]


class PFAX_PORT_INFO(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("SizeOfStruct", 0),
        NDRIntField("DeviceId", 0),
        NDRIntField("State", 0),
        NDRIntField("Flags", 0),
        NDRIntField("Rings", 0),
        NDRIntField("Priority", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("DeviceName", ""), deferred=True
        ),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("Tsid", ""), deferred=True),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("Csid", ""), deferred=True),
    ]


class FAX_ENUM_RULE_STATUS(IntEnum):
    FAX_RULE_STATUS_VALID = 0
    FAX_RULE_STATUS_EMPTY_GROUP = 1
    FAX_RULE_STATUS_ALL_GROUP_DEV_NOT_VALID = 2
    FAX_RULE_STATUS_SOME_GROUP_DEV_NOT_VALID = 3
    FAX_RULE_STATUS_BAD_DEVICE = 4


class RPC_FAX_OUTBOUND_ROUTING_RULEW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRIntField("dwAreaCode", 0),
        NDRIntField("dwCountryCode", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrCountryName", ""), deferred=True
        ),
        NDRUnionField(
            [
                (
                    NDRIntField("Destination", 0),
                    (
                        (lambda pkt: getattr(pkt, "bUseGroup", None) == 0),
                        (lambda _, val: val.tag == 0),
                    ),
                )
            ],
            NDRFullPointerField(
                NDRConfVarStrNullFieldUtf16("Destination", ""), deferred=True
            ),
            align=(4, 8),
            switch_fmt=("l", "l"),
        ),
        NDRSignedIntField("bUseGroup", 0),
    ]


class RPC_PFAX_OUTBOUND_ROUTING_RULEW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRIntField("dwAreaCode", 0),
        NDRIntField("dwCountryCode", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrCountryName", ""), deferred=True
        ),
        NDRUnionField(
            [
                (
                    NDRIntField("Destination", 0),
                    (
                        (lambda pkt: getattr(pkt, "bUseGroup", None) == 0),
                        (lambda _, val: val.tag == 0),
                    ),
                )
            ],
            NDRFullPointerField(
                NDRConfVarStrNullFieldUtf16("Destination", ""), deferred=True
            ),
            align=(4, 8),
            switch_fmt=("l", "l"),
        ),
        NDRSignedIntField("bUseGroup", 0),
    ]


class FAX_VERSION(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRSignedIntField("bValid", 0),
        NDRShortField("wMajorVersion", 0),
        NDRShortField("wMinorVersion", 0),
        NDRShortField("wMajorBuildNumber", 0),
        NDRShortField("wMinorBuildNumber", 0),
        NDRIntField("dwFlags", 0),
    ]


class PFAX_VERSION(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRSignedIntField("bValid", 0),
        NDRShortField("wMajorVersion", 0),
        NDRShortField("wMinorVersion", 0),
        NDRShortField("wMajorBuildNumber", 0),
        NDRShortField("wMinorBuildNumber", 0),
        NDRIntField("dwFlags", 0),
    ]


class FAX_OUTBOX_CONFIG(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRSignedIntField("bAllowPersonalCP", 0),
        NDRSignedIntField("bUseDeviceTSID", 0),
        NDRIntField("dwRetries", 0),
        NDRIntField("dwRetryDelay", 0),
        NDRPacketField("dtDiscountStart", FAX_TIME(), FAX_TIME),
        NDRPacketField("dtDiscountEnd", FAX_TIME(), FAX_TIME),
        NDRIntField("dwAgeLimit", 0),
        NDRSignedIntField("bBranding", 0),
    ]


class PFAX_OUTBOX_CONFIG(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRSignedIntField("bAllowPersonalCP", 0),
        NDRSignedIntField("bUseDeviceTSID", 0),
        NDRIntField("dwRetries", 0),
        NDRIntField("dwRetryDelay", 0),
        NDRPacketField("dtDiscountStart", FAX_TIME(), FAX_TIME),
        NDRPacketField("dtDiscountEnd", FAX_TIME(), FAX_TIME),
        NDRIntField("dwAgeLimit", 0),
        NDRSignedIntField("bBranding", 0),
    ]


class FAX_ACTIVITY_LOGGING_CONFIGW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRSignedIntField("bLogIncoming", 0),
        NDRSignedIntField("bLogOutgoing", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrDBPath", ""), deferred=True
        ),
    ]


class PFAX_ACTIVITY_LOGGING_CONFIGW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRSignedIntField("bLogIncoming", 0),
        NDRSignedIntField("bLogOutgoing", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrDBPath", ""), deferred=True
        ),
    ]


class FAX_PORT_INFO_EXW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRIntField("dwDeviceID", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpcwstrDeviceName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrDescription", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpcwstrProviderName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpcwstrProviderGUID", ""), deferred=True
        ),
        NDRSignedIntField("bSend", 0),
        NDRInt3264EnumField("ReceiveMode", 0, FAX_ENUM_DEVICE_RECEIVE_MODE),
        NDRIntField("dwStatus", 0),
        NDRIntField("dwRings", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrCsid", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrTsid", ""), deferred=True
        ),
    ]


class PFAX_PORT_INFO_EXW(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRIntField("dwDeviceID", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpcwstrDeviceName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrDescription", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpcwstrProviderName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpcwstrProviderGUID", ""), deferred=True
        ),
        NDRSignedIntField("bSend", 0),
        NDRInt3264EnumField("ReceiveMode", 0, FAX_ENUM_DEVICE_RECEIVE_MODE),
        NDRIntField("dwStatus", 0),
        NDRIntField("dwRings", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrCsid", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpwstrTsid", ""), deferred=True
        ),
    ]


class FAX_SERVER_ACTIVITY(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRIntField("dwIncomingMessages", 0),
        NDRIntField("dwRoutingMessages", 0),
        NDRIntField("dwOutgoingMessages", 0),
        NDRIntField("dwDelegatedOutgoingMessages", 0),
        NDRIntField("dwQueuedMessages", 0),
        NDRIntField("dwErrorEvents", 0),
        NDRIntField("dwWarningEvents", 0),
        NDRIntField("dwInformationEvents", 0),
    ]


class PFAX_SERVER_ACTIVITY(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRIntField("dwSizeOfStruct", 0),
        NDRIntField("dwIncomingMessages", 0),
        NDRIntField("dwRoutingMessages", 0),
        NDRIntField("dwOutgoingMessages", 0),
        NDRIntField("dwDelegatedOutgoingMessages", 0),
        NDRIntField("dwQueuedMessages", 0),
        NDRIntField("dwErrorEvents", 0),
        NDRIntField("dwWarningEvents", 0),
        NDRIntField("dwInformationEvents", 0),
    ]


class FAX_REASSIGN_INFO(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpcwstrRecipients", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpcwstrSenderName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpcwstrSenderFaxNumber", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpcwstrSubject", ""), deferred=True
        ),
        NDRSignedIntField("bHasCoverPage", 0),
    ]


class PFAX_REASSIGN_INFO(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpcwstrRecipients", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpcwstrSenderName", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpcwstrSenderFaxNumber", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("lpcwstrSubject", ""), deferred=True
        ),
        NDRSignedIntField("bHasCoverPage", 0),
    ]


class FAX_MESSAGE_PROPS(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [NDRIntField("dwValidityMask", 0), NDRIntField("dwMsgFlags", 0)]


class PFAX_MESSAGE_PROPS(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [NDRIntField("dwValidityMask", 0), NDRIntField("dwMsgFlags", 0)]


class FAX_JOB_ENTRY(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("SizeOfStruct", 0),
        NDRIntField("JobId", 0),
        NDRFullPointerField(NDRShortField("UserName", 0), deferred=True),
        NDRIntField("JobType", 0),
        NDRIntField("QueueStatus", 0),
        NDRIntField("Status", 0),
        NDRIntField("Size", 0),
        NDRIntField("PageCount", 0),
        NDRFullPointerField(NDRShortField("RecipientNumber", 0), deferred=True),
        NDRFullPointerField(NDRShortField("RecipientName", 0), deferred=True),
        NDRFullPointerField(NDRShortField("Tsid", 0), deferred=True),
        NDRFullPointerField(NDRShortField("SenderName", 0), deferred=True),
        NDRFullPointerField(NDRShortField("SenderCompany", 0), deferred=True),
        NDRFullPointerField(NDRShortField("SenderDept", 0), deferred=True),
        NDRFullPointerField(NDRShortField("BillingCode", 0), deferred=True),
        NDRIntField("ScheduleAction", 0),
        NDRPacketField("ScheduleTime", SYSTEMTIME(), SYSTEMTIME),
        NDRIntField("DeliveryReportType", 0),
        NDRFullPointerField(NDRShortField("DeliveryReportAddress", 0), deferred=True),
        NDRFullPointerField(NDRShortField("DocumentName", 0), deferred=True),
    ]


class PFAX_JOB_ENTRY(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("SizeOfStruct", 0),
        NDRIntField("JobId", 0),
        NDRFullPointerField(NDRShortField("UserName", 0), deferred=True),
        NDRIntField("JobType", 0),
        NDRIntField("QueueStatus", 0),
        NDRIntField("Status", 0),
        NDRIntField("Size", 0),
        NDRIntField("PageCount", 0),
        NDRFullPointerField(NDRShortField("RecipientNumber", 0), deferred=True),
        NDRFullPointerField(NDRShortField("RecipientName", 0), deferred=True),
        NDRFullPointerField(NDRShortField("Tsid", 0), deferred=True),
        NDRFullPointerField(NDRShortField("SenderName", 0), deferred=True),
        NDRFullPointerField(NDRShortField("SenderCompany", 0), deferred=True),
        NDRFullPointerField(NDRShortField("SenderDept", 0), deferred=True),
        NDRFullPointerField(NDRShortField("BillingCode", 0), deferred=True),
        NDRIntField("ScheduleAction", 0),
        NDRPacketField("ScheduleTime", SYSTEMTIME(), SYSTEMTIME),
        NDRIntField("DeliveryReportType", 0),
        NDRFullPointerField(NDRShortField("DeliveryReportAddress", 0), deferred=True),
        NDRFullPointerField(NDRShortField("DocumentName", 0), deferred=True),
    ]



### ms-tsts_rcmpublic.idl
"""
RPC definitions for the following interfaces:
- RCMPublic (v1.0): bde95fdf-eee0-45de-9e12-e5a61cd0d4fe
- RCMListener (v1.0): 497d95a6-2d27-4bf5-9bbd-a6046957133c
This file is auto-generated by midl-to-scapy, do not modify.
"""

from enum import IntEnum
import uuid

from scapy.fields import StrFixedLenField, StrFixedLenFieldUtf16
from scapy.layers.dcerpc import (
    NDRPacket,
    DceRpcOp,
    NDRConfPacketListField,
    NDRConfStrLenField,
    NDRConfVarStrLenField,
    NDRConfVarStrNullField,
    NDRConfVarStrNullFieldUtf16,
    NDRContextHandle,
    NDRInt3264EnumField,
    NDRIntField,
    NDRPacketField,
    NDRShortField,
    NDRSignedIntField,
    NDRSignedLongField,
    NDRUnionField,
    register_dcerpc_interface,
)


class RpcGetClientData_Request(NDRPacket):
    fields_desc = [NDRIntField("SessionId", 0)]


class RpcGetClientData_Response(NDRPacket):
    fields_desc = [
        NDRConfStrLenField("ppBuff", "", size_is=lambda pkt: pkt.pOutBuffByteLen),
        NDRIntField("pOutBuffByteLen", None, size_of="ppBuff"),
        NDRIntField("status", 0),
    ]


class RpcGetConfigData_Request(NDRPacket):
    fields_desc = [NDRIntField("SessionId", 0)]


class RpcGetConfigData_Response(NDRPacket):
    fields_desc = [
        NDRConfStrLenField("ppBuff", "", size_is=lambda pkt: pkt.pOutBuffByteLen),
        NDRIntField("pOutBuffByteLen", None, size_of="ppBuff"),
        NDRIntField("status", 0),
    ]


class PROTOCOLSTATUS_INFO_TYPE(IntEnum):
    PROTOCOLSTATUS_INFO_BASIC = 0
    PROTOCOLSTATUS_INFO_EXTENDED = 1


class RpcGetProtocolStatus_Request(NDRPacket):
    fields_desc = [
        NDRIntField("SessionId", 0),
        NDRInt3264EnumField("InfoType", 0, PROTOCOLSTATUS_INFO_TYPE),
    ]


class RpcGetProtocolStatus_Response(NDRPacket):
    fields_desc = [
        NDRConfStrLenField("ppProtoStatus", "", size_is=lambda pkt: pkt.pcbProtoStatus),
        NDRIntField("pcbProtoStatus", None, size_of="ppProtoStatus"),
        NDRIntField("status", 0),
    ]


class RpcGetLastInputTime_Request(NDRPacket):
    fields_desc = [NDRIntField("SessionId", 0)]


class RpcGetLastInputTime_Response(NDRPacket):
    fields_desc = [NDRSignedLongField("pLastInputTime", 0), NDRIntField("status", 0)]


class ipv4_sub0(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRShortField("sin_port", 0),
        NDRIntField("in_addr", 0),
        StrFixedLenField("sin_zero", "", length=8),
    ]


class ipv6_sub1(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRShortField("sin6_port", 0),
        NDRIntField("sin6_flowinfo", 0),
        StrFixedLenFieldUtf16("sin6_addr", "", length=8 * 2),
        NDRIntField("sin6_scope_id", 0),
    ]


class PRCM_REMOTEADDRESS(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRShortField("sin_family", 0),
        NDRUnionField(
            [
                (
                    NDRPacketField("u", ipv4_sub0(), ipv4_sub0),
                    ((lambda pkt: None == 2), (lambda _, val: val.tag == 2)),
                ),
                (
                    NDRPacketField("u", ipv6_sub1(), ipv6_sub1),
                    ((lambda pkt: None == 23), (lambda _, val: val.tag == 23)),
                ),
            ],
            StrFixedLenField("u", "", length=0),
            align=(4, 4),
            switch_fmt=("I", "I"),
        ),
    ]


class RpcGetRemoteAddress_Request(NDRPacket):
    fields_desc = [NDRIntField("SessionId", 0)]


class RpcGetRemoteAddress_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("pRemoteAddress", PRCM_REMOTEADDRESS(), PRCM_REMOTEADDRESS),
        NDRIntField("status", 0),
    ]


class LISTENERENUM_LEVEL1(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRSignedIntField("Id", 0),
        NDRSignedIntField("bListening", 0),
        StrFixedLenFieldUtf16("Name", "", length=33 * 2),
    ]


class PLISTENERENUM(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRIntField("Level", 0),
        NDRUnionField(
            [
                (
                    NDRPacketField("Data", LISTENERENUM_LEVEL1(), LISTENERENUM_LEVEL1),
                    (
                        (lambda pkt: getattr(pkt, "Level", None) == 1),
                        (lambda _, val: val.tag == 1),
                    ),
                )
            ],
            StrFixedLenField("Data", "", length=0),
            align=(4, 4),
            switch_fmt=("L", "L"),
        ),
    ]


class RpcGetAllListeners_Request(NDRPacket):
    fields_desc = [NDRIntField("Level", 0)]


class RpcGetAllListeners_Response(NDRPacket):
    fields_desc = [
        NDRConfPacketListField(
            "ppListeners",
            [],
            PLISTENERENUM,
            size_is=lambda pkt: pkt.pNumListeners,
            ptr_pack=True,
        ),
        NDRIntField("pNumListeners", None, size_of="ppListeners"),
        NDRIntField("status", 0),
    ]


class RpcGetSessionProtocolLastInputTime_Request(NDRPacket):
    fields_desc = [
        NDRIntField("SessionId", 0),
        NDRInt3264EnumField("InfoType", 0, PROTOCOLSTATUS_INFO_TYPE),
    ]


class RpcGetSessionProtocolLastInputTime_Response(NDRPacket):
    fields_desc = [
        NDRConfStrLenField("ppProtoStatus", "", size_is=lambda pkt: pkt.pcbProtoStatus),
        NDRIntField("pcbProtoStatus", None, size_of="ppProtoStatus"),
        NDRSignedLongField("pLastInputTime", 0),
        NDRIntField("status", 0),
    ]


class RpcGetUserCertificates_Request(NDRPacket):
    fields_desc = [NDRIntField("SessionId", 0)]


class RpcGetUserCertificates_Response(NDRPacket):
    fields_desc = [
        NDRIntField("pcCerts", 0),
        NDRConfStrLenField("ppbCerts", "", size_is=lambda pkt: pkt.pcbCerts),
        NDRIntField("pcbCerts", None, size_of="ppbCerts"),
        NDRIntField("status", 0),
    ]


class QUERY_SESSION_DATA_TYPE(IntEnum):
    QUERY_SESSION_DATA_MODULE = 0
    QUERY_SESSION_DATA_WDCONFIG = 1
    QUERY_SESSION_DATA_VIRTUALDATA = 2
    QUERY_SESSION_DATA_LICENSE = 3
    QUERY_SESSION_DATA_DEVICEID = 4
    QUERY_SESSION_DATA_LICENSE_VALIDATION = 5


class RpcQuerySessionData_Request(NDRPacket):
    fields_desc = [
        NDRIntField("SessionId", 0),
        NDRInt3264EnumField("type", 0, QUERY_SESSION_DATA_TYPE),
        NDRConfStrLenField("pbInputData", "", size_is=lambda pkt: pkt.cbInputData),
        NDRIntField("cbInputData", None, size_of="pbInputData"),
        NDRIntField("cbSessionData", 0),
    ]


class RpcQuerySessionData_Response(NDRPacket):
    fields_desc = [
        NDRConfVarStrLenField(
            "pbSessionData",
            "",
            size_is=lambda pkt: pkt.cbSessionData,
            length_is=lambda pkt: pkt.pcbReturnLength,
        ),
        NDRIntField("pcbReturnLength", None, size_of="pbSessionData"),
        NDRIntField("pcbRequireBufferSize", 0),
        NDRIntField("status", 0),
    ]


RCMPUBLIC_OPNUMS = {
    0: DceRpcOp(RpcGetClientData_Request, RpcGetClientData_Response),
    1: DceRpcOp(RpcGetConfigData_Request, RpcGetConfigData_Response),
    2: DceRpcOp(RpcGetProtocolStatus_Request, RpcGetProtocolStatus_Response),
    3: DceRpcOp(RpcGetLastInputTime_Request, RpcGetLastInputTime_Response),
    4: DceRpcOp(RpcGetRemoteAddress_Request, RpcGetRemoteAddress_Response),
    # 5: Opnum5NotUsedOnWire,
    # 6: Opnum6NotUsedOnWire,
    # 7: Opnum7NotUsedOnWire,
    8: DceRpcOp(RpcGetAllListeners_Request, RpcGetAllListeners_Response),
    9: DceRpcOp(
        RpcGetSessionProtocolLastInputTime_Request,
        RpcGetSessionProtocolLastInputTime_Response,
    ),
    10: DceRpcOp(RpcGetUserCertificates_Request, RpcGetUserCertificates_Response),
    11: DceRpcOp(RpcQuerySessionData_Request, RpcQuerySessionData_Response),
}
register_dcerpc_interface(
    name="RCMPublic",
    uuid=uuid.UUID("bde95fdf-eee0-45de-9e12-e5a61cd0d4fe"),
    version="1.0",
    opnums=RCMPUBLIC_OPNUMS,
)


class RpcOpenListener_Request(NDRPacket):
    fields_desc = [NDRConfVarStrNullFieldUtf16("szListenerName", "")]


class RpcOpenListener_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("phListener", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class RpcCloseListener_Request(NDRPacket):
    fields_desc = [NDRPacketField("phListener", NDRContextHandle(), NDRContextHandle)]


class RpcCloseListener_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("phListener", NDRContextHandle(), NDRContextHandle),
        NDRIntField("status", 0),
    ]


class RpcStopListener_Request(NDRPacket):
    fields_desc = [NDRPacketField("hListener", NDRContextHandle(), NDRContextHandle)]


class RpcStopListener_Response(NDRPacket):
    fields_desc = [NDRIntField("status", 0)]


class RpcStartListener_Request(NDRPacket):
    fields_desc = [NDRPacketField("hListener", NDRContextHandle(), NDRContextHandle)]


class RpcStartListener_Response(NDRPacket):
    fields_desc = [NDRIntField("status", 0)]


class RpcIsListening_Request(NDRPacket):
    fields_desc = [NDRPacketField("hListener", NDRContextHandle(), NDRContextHandle)]


class RpcIsListening_Response(NDRPacket):
    fields_desc = [NDRSignedIntField("pbIsListening", 0), NDRIntField("status", 0)]


RCMLISTENER_OPNUMS = {
    0: DceRpcOp(RpcOpenListener_Request, RpcOpenListener_Response),
    1: DceRpcOp(RpcCloseListener_Request, RpcCloseListener_Response),
    2: DceRpcOp(RpcStopListener_Request, RpcStopListener_Response),
    3: DceRpcOp(RpcStartListener_Request, RpcStartListener_Response),
    4: DceRpcOp(RpcIsListening_Request, RpcIsListening_Response),
}
register_dcerpc_interface(
    name="RCMListener",
    uuid=uuid.UUID("497d95a6-2d27-4bf5-9bbd-a6046957133c"),
    version="1.0",
    opnums=RCMLISTENER_OPNUMS,
)
